"""Markov chain transition operators."""

import logging
import numpy as np
from mici.utils import LogRepFloat
from mici.errors import (
    IntegratorError, NonReversibleStepError, ConvergenceError)

logger = logging.getLogger(__name__)


class _AbstractMomentumTransition(object):
    """Base class for momentum transitions."""

    def __init__(self, system):
        """
        Args:
            system: Hamiltonian system to be simulated.
        """
        self.system = system
        self.state_variables = {'mom'}


class IndependentMomentumTransition(_AbstractMomentumTransition):
    """Independent momentum transition.

    Independently resamples the momentum component of the state from its
    conditional distribution given the remaining state.
    """

    def sample(self, state, rng):
        state.mom = self.system.sample_momentum(state, rng)
        return state, None


class CorrelatedMomentumTransition(_AbstractMomentumTransition):
    """Correlated (partial) momentum transition.

    Rather than independently sampling a new momentum, instead a pertubative
    Crank-Nicolson type update which produces a new momentum value with a
    specified correlation with the previous value is used. It is assumed that
    the conditional distribution of the momenta is zero-mean Gaussian such that
    the Crank-Nicolson update leaves the momenta conditional distribution
    exactly invariant. This approach is sometimes known as partial momentum
    refreshing or updating, and was originally proposed in [1].

    If the resampling coefficient is equal to zero then the momentum is not
    randomised at all and succesive applications of the coupled integration
    transitions will continue along the same simulated Hamiltonian trajectory.
    When an integration transition is accepted this means the subsequent
    simulated trajectory will continue evolving in the same direction and so
    not randomising the momentum will reduce random-walk behaviour. However on
    a rejection the integration direction is reversed and so without
    randomisation the trajectory will exactly backtrack along the previous
    tractory states. A resampling coefficient of one corresponds to the
    standard case of independent resampling of the momenta while intermediate
    values between zero and one correspond to varying levels of correlation
    between the pre and post update momentums.

    References:

      1. Horowitz, A.M., 1991. A generalized guided Monte Carlo algorithm.
         Phys. Lett. B, 268(CERN-TH-6172-91), pp.247-252.
    """

    def __init__(self, system, mom_resample_coeff=1.):
        super().__init__(system)
        self.mom_resample_coeff = mom_resample_coeff

    def sample(self, state, rng):
        if self.mom_resample_coeff == 1:
            state.mom = self.system.sample_momentum(state, rng)
        elif self.mom_resample_coeff != 0:
            mom_ind = self.system.sample_momentum(state, rng)
            state.mom *= (1. - self.mom_resample_coeff**2)**0.5
            state.mom += self.mom_resample_coeff * mom_ind
        return state, None


class _AbstractIntegrationTransition(object):

    def __init__(self, system, integrator):
        """
        Args:
            system: Hamiltonian system to be simulated.
            integrator: Symplectic integrator appropriate to the specified
                Hamiltonian system.
        """
        self.system = system
        self.integrator = integrator
        self.statistic_types = {
            'hamiltonian': (np.float64, np.nan),
            'n_step': (np.int64, -1),
            'accept_prob': (np.float64, np.nan),
            'non_reversible_step': (np.bool, False),
            'convergence_error': (np.bool, False)
        }
        self.state_variables = {'pos', 'mom', 'dir'}


class _AbstractMetropolisIntegrationTransition(_AbstractIntegrationTransition):
    """Base for HMC methods using a Metropolis accept step to sample new state.

    In each transition a trajectory is generated by integrating the Hamiltonian
    dynamics from the current state in the current integration time direction
    for a number of integrator steps.

    The state at the end of the trajectory with the integration direction
    negated (this ensuring the proposed move is an involution) is used as the
    proposal in a Metropolis acceptance step. The integration direction is then
    deterministically negated again irrespective of the accept decision, with
    the effect being that on acceptance the integration direction will be equal
    to its initial value and on rejection the integration direction will be
    the negation of its initial value.
    """

    def _sample_n_step(self, state, n_step, rng):
        h_init = self.system.h(state)
        state_p = state
        try:
            for s in range(n_step):
                state_p = self.integrator.step(state_p)
        except IntegratorError as e:
            logger.info(
                f'Terminating trajectory due to integrator error:\n{e!s}')
            return state, {
                'hamiltonian': h_init, 'accept_prob': 0, 'n_step': s,
                'non_reversible_step': isinstance(e, NonReversibleStepError),
                'convergence_error': isinstance(e, ConvergenceError)}
        state_p.dir *= -1
        h_final = self.system.h(state_p)
        metrop_ratio = np.exp(h_init - h_final)
        accept_prob = 0 if np.isnan(metrop_ratio) else min(1, metrop_ratio)
        if rng.uniform() < accept_prob:
            state = state_p
        state.dir *= -1
        stats = {'hamiltonian': self.system.h(state),
                 'accept_prob': accept_prob, 'n_step': n_step,
                 'non_reversible_step': False, 'convergence_error': False}
        return state, stats


class MetropolisStaticIntegrationTransition(
        _AbstractMetropolisIntegrationTransition):
    """Static integration transition with Metropolis sampling of new state.

    In this variant the trajectory is generated by integrating the state
    through time a fixed number of integrator steps. This is original proposed
    Hybrid Monte Carlo (often now instead termed Hamiltonian Monte Carlo)
    algorithm [1,2].

    References:

      1. Duane, S., Kennedy, A.D., Pendleton, B.J. and Roweth, D., 1987.
         Hybrid Monte Carlo. Physics letters B, 195(2), pp.216-222.
      2. Neal, R.M., 2011. MCMC using Hamiltonian dynamics.
         Handbook of Markov Chain Monte Carlo, 2(11), p.2.
    """

    def __init__(self, system, integrator, n_step):
        super().__init__(system, integrator)
        self.n_step = n_step

    def sample(self, state, rng):
        return self._sample_n_step(state, self.n_step, rng)


class MetropolisRandomIntegrationTransition(
        _AbstractMetropolisIntegrationTransition):
    """Random integration transition with Metropolis sampling of new state.

    In each transition a trajectory is generated by integrating the state in
    the current integration direction in time a random integer number of
    integrator steps sampled from the uniform distribution on an integer
    interval. The randomisation of the number of integration steps avoids the
    potential of the chain mixing poorly due to using an integration time close
    to the period of (near) periodic systems [1,2].

    References:

      1. Neal, R.M., 2011. MCMC using Hamiltonian dynamics.
         Handbook of Markov Chain Monte Carlo, 2(11), p.2.
      2. Mackenzie, P.B., 1989. An improved hybrid Monte Carlo method.
         Physics Letters B, 226(3-4), pp.369-371.
    """

    def __init__(self, system, integrator, n_step_range):
        super().__init__(system, integrator)
        n_step_lower, n_step_upper = n_step_range
        assert n_step_lower > 0 and n_step_lower < n_step_upper
        self.n_step_range = n_step_range

    def sample(self, state, rng):
        n_step = rng.random_integers(*self.n_step_range)
        return self._sample_n_step(state, n_step, rng)


def euclidean_no_u_turn_criterion(system, state_1, state_2, sum_mom):
    """No-U-turn termination criterion for Euclidean manifolds [1].

    Terminates trajectories when the velocities at the terminal states of
    the trajectory both have negative dot products with the vector from
    the position of the first terminal state to the position of the second
    terminal state, corresponding to further evolution of the trajectory
    reducing the distance between the terminal state positions.

    Args:
        system (HamiltonianSystem): Hamiltonian system being integrated.
        state_1 (ChainState): First terminal state of trajectory.
        state_2 (ChainState): Second terminal state of trajectory.
        sum_mom (array): Sum of momentums of trajectory states.

    Returns:
        True if termination criterion is satisfied.

    References:

      1. Hoffman, M.D. and Gelman, A., 2014. The No-U-turn sampler:
         adaptively setting path lengths in Hamiltonian Monte Carlo.
         Journal of Machine Learning Research, 15(1), pp.1593-1623.
    """
    return (
        np.sum(system.dh_dmom(state_1) * (state_2.pos - state_1.pos)) < 0 or
        np.sum(system.dh_dmom(state_2) * (state_2.pos - state_1.pos)) < 0)


def riemannian_no_u_turn_criterion(system, state_1, state_2, sum_mom):
    """Generalised no-U-turn termination criterion on Riemannian manifolds [2].

    Terminates trajectories when the velocities at the terminal states of
    the trajectory both have negative dot products with the sum of the
    the momentums across the trajectory from the first to second terminal state
    of the first terminal state to the position of the second terminal state.
    This generalises the no-U-turn criterion of [1] to Riemannian manifolds
    where due to the intrinsic curvature of the space the geodesic between
    two points is general no longer a straight line.

    Args:
        system (HamiltonianSystem): Hamiltonian system being integrated.
        state_1 (ChainState): First terminal state of trajectory.
        state_2 (ChainState): Second terminal state of trajectory.
        sum_mom (array): Sum of momentums of trajectory states.

    Returns:
        True if termination criterion is satisfied.

    References:

      1. Hoffman, M.D. and Gelman, A., 2014. The No-U-turn sampler:
         adaptively setting path lengths in Hamiltonian Monte Carlo.
         Journal of Machine Learning Research, 15(1), pp.1593-1623.
      2. Betancourt, M., 2013. Generalizing the no-U-turn sampler to Riemannian
         manifolds. arXiv preprint arXiv:1304.1920.
    """
    return (
        np.sum(system.dh_dmom(state_1) * sum_mom) < 0 or
        np.sum(system.dh_dmom(state_2) * sum_mom) < 0)


class MultinomialDynamicIntegrationTransition(_AbstractIntegrationTransition):
    """Dynamic integration transition with multinomial sampling of new state.

    In each transition a binary tree of states is recursively computed by
    integrating randomly forward and backward in time by a number of steps
    equal to the previous tree size [1,2] until a termination criteria on the
    tree leaves is met. The next chain state is chosen from the candidate
    states using a progressive multinomial sampling scheme [2] based on the
    relative probability densities of the different candidate states, with the
    resampling biased towards states further from the current state.

    References:

      1. Hoffman, M.D. and Gelman, A., 2014. The No-U-turn sampler:
         adaptively setting path lengths in Hamiltonian Monte Carlo.
         Journal of Machine Learning Research, 15(1), pp.1593-1623.
      2. Betancourt, M., 2017. A conceptual introduction to Hamiltonian Monte
         Carlo. arXiv preprint arXiv:1701.02434.
    """

    def __init__(self, system, integrator,
                 max_tree_depth=10, max_delta_h=1000,
                 termination_criterion=riemannian_no_u_turn_criterion):
        super().__init__(system, integrator)
        self.max_tree_depth = max_tree_depth
        self.max_delta_h = max_delta_h
        self._termination_criterion = termination_criterion
        self.statistic_types['tree_depth'] = (np.int64, -1)
        self.statistic_types['diverging'] = (np.bool, False)

    def termination_criterion(self, state_1, state_2, sum_mom):
        return self._termination_criterion(
            self.system, state_1, state_2, sum_mom)

    # Key to subscripts used in build_tree and sample_dynamics_transition
    # _p : proposal
    # _n : next
    # _l : left (negative direction)
    # _r : right (positive direction)
    # _s : subtree
    # _i : inner subsubtree
    # _o : outer subsubtree

    def _build_tree(
            self, depth, state, sum_mom, sum_weight, stats, h_init, rng):
        if depth == 0:
            # recursion base case
            try:
                state = self.integrator.step(state)
                h = self.system.h(state)
                h = np.inf if np.isnan(h) else h
                sum_mom += state.mom
                sum_weight += LogRepFloat(log_val=-h)
                stats['sum_acc_prob'] += min(1, np.exp(h_init - h))
                stats['n_step'] += 1
                terminate = h - h_init > self.max_delta_h
                if terminate:
                    stats['diverging'] = True
                    logger.info(
                        f'Terminating build_tree due to integrator divergence '
                        f'(delta_h = {h - h_init:.1e}).')
            except IntegratorError as e:
                logger.info(
                    f'Terminating build_tree due to integrator error:\n{e!s}')
                stats['non_reversible_step'] = isinstance(
                    e, NonReversibleStepError)
                stats['convergence_error'] = isinstance(e, ConvergenceError)
                state = None
                terminate = True
            return terminate, state, state, state
        sum_mom_i, sum_mom_o = np.zeros((2,) + state.mom.shape)
        sum_weight_i, sum_weight_o = LogRepFloat(0.), LogRepFloat(0.)
        # build inner subsubtree
        terminate_i, state_i, state, state_pi = self._build_tree(
            depth - 1, state, sum_mom_i, sum_weight_i, stats, h_init, rng)
        if terminate_i:
            return True, None, None, None
        # build outer subsubtree
        terminate_o, _, state_o, state_po = self._build_tree(
            depth - 1, state, sum_mom_o, sum_weight_o, stats, h_init, rng)
        if terminate_o:
            return True, None, None, None
        # independently sample proposal from 2 subsubtrees by relative weights
        sum_weight_s = sum_weight_i + sum_weight_o
        accept_o_prob = sum_weight_o / sum_weight_s
        state_p = state_po if rng.uniform() < accept_o_prob else state_pi
        # update overall tree weight
        sum_weight += sum_weight_s
        # calculate termination criteria for subtree
        sum_mom_s = sum_mom_i + sum_mom_o
        terminate_s = self.termination_criterion(state_i, state_o, sum_mom_s)
        # update overall tree summed momentum
        sum_mom += sum_mom_s
        return terminate_s, state_i, state_o, state_p

    def sample(self, state, rng):
        h_init = self.system.h(state)
        sum_mom = state.mom.copy()
        sum_weight = LogRepFloat(log_val=-h_init)
        stats = {'n_step': 0, 'sum_acc_prob': 0.}
        state_n, state_l, state_r = state, state.copy(), state.copy()
        # set integration directions of initial left and right tree leaves
        state_l.dir = -1
        state_r.dir = +1
        for depth in range(self.max_tree_depth):
            # uniformly sample direction to expand tree in
            direction = 2 * (rng.uniform() < 0.5) - 1
            sum_mom_s = np.zeros(state.mom.shape)
            sum_weight_s = LogRepFloat(0.)
            if direction == 1:
                # expand tree by adding subtree to right edge
                terminate_s, _, state_r, state_p = self._build_tree(
                    depth, state_r, sum_mom_s, sum_weight_s, stats, h_init,
                    rng)
            else:
                # expand tree by adding subtree to left edge
                terminate_s, _, state_l, state_p = self._build_tree(
                    depth, state_l, sum_mom_s, sum_weight_s, stats, h_init,
                    rng)
            if terminate_s:
                break
            # progressively sample new state by choosing between
            # current new state and proposal from new subtree, biasing
            # towards the new subtree proposal
            if rng.uniform() < sum_weight_s / sum_weight:
                state_n = state_p
            sum_weight += sum_weight_s
            sum_mom += sum_mom_s
            if self.termination_criterion(state_l, state_r, sum_mom):
                break
        if stats['n_step'] > 0:
            stats['accept_prob'] = stats['sum_acc_prob'] / stats['n_step']
        else:
            stats['accept_prob'] = 0.
        stats['hamiltonian'] = self.system.h(state_n)
        stats['tree_depth'] = depth
        return state_n, stats



"""Automatic differentation fallback for constructing derivative functions.

Currently only Autograd is available as a backend but other backend
frameworks may be available as alternative fallbacks in the future.
"""

AUTOGRAD_AVAILABLE = True
try:
    import mici.autograd_wrapper as autograd_wrapper
except ImportError:
    AUTOGRAD_AVAILABLE = False


"""List of names of valid differential operators.

Any automatic differentiation framework wrapper module will need to provide
all of these operators as callables (with a single function as argument) to
fully support all of the required derivative functions.
"""
DIFF_OPS = [
    # vector Jacobian product and value
    'vjp_and_value',
    # gradient and value for scalar valued functions
    'grad_and_value',
    # Hessian matrix, gradient and value for scalar valued functions
    'hessian_grad_and_value',
    # matrix Tressian product, gradient and value for scalar valued
    # functions
    'mtp_hessian_grad_and_value',
    # Jacobian matrix and value for vector valued functions
    'jacobian_and_value',
    # matrix Hessian product, Jacobian matrix and value for vector valued
    # functions
    'mhp_jacobian_and_value',
]


def autodiff_fallback(diff_func, func, diff_op_name, name):
    """Generate derivative function automatically if not provided.

    Uses automatic differentiation to generate a function corresponding to a
    differential operator applied to a function if an alternative
    implementation of the derivative function has not been provided.

    Args:
        diff_func (callable or None): Either a callable implementing the
            required derivative function or `None` if none was provided.
        func (callable): Function to differentiate.
        diff_op_name (string): String specifying name of differential operator
            from automatic differentiation framework wrapper to use to generate
            required derivative function.
        name (string): Name of derivative function to use in error message.

    Returns:
        `diff_func` value if not `None` otherwise generated derivative of
        `func` by applying named differential operator.
    """
    if diff_func is not None:
        return diff_func
    elif diff_op_name not in DIFF_OPS:
        raise ValueError(
            f'Differential operator {diff_op_name} is not defined.')
    elif AUTOGRAD_AVAILABLE:
        return getattr(autograd_wrapper, diff_op_name)(func)
    elif not AUTOGRAD_AVAILABLE:
        raise ValueError(
            f'Autograd not available therefore {name} must be provided.')



"""Objects for recording state of a Markov chain."""

import copy


def cache_in_state(*depends_on):
    """Decorator to memoize / cache output of a function of state variable(s).

    Used to wrap functions of a chain state vaiable(s) to allow caching of
    the values computed to prevent recomputation when possible.

    Args:
       *depends_on: One or more strings defining which state variables the
           computed values depend on e.g. 'pos', 'mom', such that the cache is
           correctly cleared when one of these parent dependency's value
           changes.
    """
    def cache_in_state_decorator(method):
        def wrapper(self, state):
            key = (type(self).__name__ + '.' + method.__name__, id(self))
            if key not in state._cache:
                for dep in depends_on:
                    state._dependencies[dep].add(key)
            if key not in state._cache or state._cache[key] is None:
                state._cache[key] = method(self, state)
            return state._cache[key]
        return wrapper
    return cache_in_state_decorator


def multi_cache_in_state(depends_on, vars, primary_index=0):
    """Decorator to cache multiple outputs of a function of state variable(s).

    Used to wrap functions of a chain state vaiable(s) to allow caching of
    the values computed to prevent recomputation when possible.

    This variant allows for functions which also cache intermediate computed
    results which may be used separately elsewhere for example the value of a
    function calculate in the forward pass of a reverse-mode automatic-
    differentation implementation of its gradient.

    Args:
        depends_on: a list of strings defining which state variables the
            computed values depend on e.g. ['pos', 'mom'], such that the cache
            is correctly cleared when one of these parent dependency's value
            changes.
        vars: a list of strings defining the variables in the state cache dict
            corresponding to the outputs of the wrapped function (method) in
            the corresponding returned order.
        primary_index: index of primary output of function (i.e. value to be
            returned) in vars list / position in output of function.
    """
    def multi_cache_in_state_decorator(method):
        def wrapper(self, state):
            id_ = id(self)
            type_prefix = type(self).__name__ + '.'
            prim_key = (type_prefix + vars[primary_index], id_)
            keys = [(type_prefix + v, id_) for v in vars]
            for i, key in enumerate(keys):
                if key not in state._cache:
                    for dep in depends_on:
                        state._dependencies[dep].add(key)
            if prim_key not in state._cache or state._cache[prim_key] is None:
                vals = method(self, state)
                if isinstance(vals, tuple):
                    for k, v in zip(keys, vals):
                        state._cache[k] = v
                else:
                    state._cache[prim_key] = vals
            return state._cache[prim_key]
        return wrapper
    return multi_cache_in_state_decorator


class ChainState(object):
    """Markov chain state.

    As well as recording the chain state variable values, the state object is
    also used to cache derived quantities to avoid recalculation if these
    values are subsequently reused.
    """

    def __init__(self, _dependencies=None, _cache=None, **vars):
        """Create a new `ChainState` instance.

        Any keyword arguments passed to the constructor will be used to set
        state variable attributes of state object for example

            state = ChainState(pos=pos_val, mom=mom_val, dir=dir_val)

        will return a `ChainState` instance `state` with variable attributes
        `state.pos`, `state.mom` and `state.dir` with initial values set to
        `pos_val`, `mom_val` and `dir_val` respectively. The keyword arguments
        `_dependencies` and `_cache` are reserved for the dependency set and
        cache dictionary respectively used to implement the caching of derived
        quantities and cannot be used as state variable names.
        """
        # set vars attribute by directly writing to __dict__ to ensure set
        # before any cally to __setattr__
        self.__dict__['vars'] = vars
        if _dependencies is None:
            _dependencies = {name: set() for name in vars}
        self._dependencies = _dependencies
        if _cache is None:
            _cache = {}
        self._cache = _cache

    def __getattr__(self, name):
        if name in self.vars:
            return self.vars[name]
        else:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        if name in self.vars:
            self.vars[name] = value
            # clear any dependent cached values
            for dep in self._dependencies[name]:
                self._cache[dep] = None
        else:
            return super().__setattr__(name, value)

    def __contains__(self, name):
        return name in self.vars

    def copy(self):
        """Create a deep copy of the state object.

        Returns:
            A copy of the state object which can be updated without
            affecting the original object's attributes.
        """
        return type(self)(
            _cache=self._cache.copy(), _dependencies=self._dependencies,
            **{name: copy.copy(val) for name, val in self.vars.items()})

    def __str__(self):
        return (
            '(\n ' +
            ',\n '.join([f'{name}={val}' for name, val in self.vars.items()]) +
            ')'
        )

    def __repr__(self):
        return type(self).__name__ + str(self)

    def __getstate__(self):
        return self.vars

    def __setstate__(self, state):
        self.__dict__['vars'] = state
        self._dependencies = {name: set() for name in self.vars}
        self._cache = {}



"""Structured matrix classes implementing basic linear algebra operations."""

import numpy as np
import numpy.linalg as nla
import scipy.linalg as sla


class _AbstractMatrix(object):

    __array_priority__ = 1

    def __init__(self, shape):
        self.shape = shape

    def __array__(self):
        if hasattr(self, 'array'):
            return self.array
        else:
            raise NotImplementedError()

    def __mul__(self, other):
        if np.isscalar(other) or np.ndim(other) == 0:
            if other == 0:
                raise NotImplementedError(
                    'Scalar multiplication by zero not implemented.')
            return self._scalar_multiply(other)
        else:
            return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if np.isscalar(other) or np.ndim(other) == 0:
            if other == 0:
                raise NotImplementedError(
                    'Scalar division by zero not implemented.')
            return self._scalar_multiply(1 / other)
        else:
            return NotImplemented

    def __neg__(self):
        return self._scalar_multiply(-1)

    def __matmul__(self, other):
        if self.shape[1] is not None and other.shape[0] != self.shape[1]:
            raise ValueError(
                f'Inconsistent dimensions for matrix multiplication: '
                f'{self.shape} and {other.shape}.')
        try:
            return self._left_matrix_multiply(other)
        except NotImplementedError:
            return MatrixProduct(self, other)

    def __rmatmul__(self, other):
        if self.shape[0] is not None and other.shape[-1] != self.shape[0]:
            raise ValueError(
                f'Inconsistent dimensions for matrix multiplication: '
                f'{other.shape} and {self.shape}.')
        return self._right_matrix_multiply(other)

    @property
    def array(self):
        return self._array

    def _left_matrix_multiply(self, other):
        return self.array @ other

    def _right_matrix_multiply(self, other):
        return other @ self.array

    def _scalar_multiply(self, other):
        raise NotImplementedError()


class _AbstractSquareMatrix(_AbstractMatrix):

    def __init__(self, size):
        super().__init__((size, size))

    @property
    def log_abs_det_sqrt(self):
        return 0.5 * self.log_abs_det

    @property
    def diagonal(self):
        return self.array.diagonal()


class _AbstractSymmetricMatrix(_AbstractSquareMatrix):

    def __init__(self, size, is_posdef=False, is_negdef=False):
        self._eigval = None
        self._eigvec = None
        self._is_posdef = is_posdef
        self._is_negdef = is_negdef
        super().__init__(size)

    def _compute_eigendecomposition(self):
        self._eigval, eigvec = nla.eigh(self.array)
        self._eigvec = OrthogonalMatrix(eigvec)

    @property
    def eigval(self):
        if self._eigval is None or self._eigvec is None:
            self._compute_eigendecomposition()
        return self._eigval

    @property
    def eigvec(self):
        if self._eigval is None or self._eigvec is None:
            self._compute_eigendecomposition()
        return self._eigvec

    @property
    def T(self):
        return self

    @property
    def log_abs_det(self):
        return np.log(np.abs(self.eigval)).sum()

    @property
    def is_posdef(self):
        return self._is_posdef

    @property
    def is_negdef(self):
        return self._is_negdef


class IdentityMatrix(_AbstractSymmetricMatrix):

    def __init__(self, size=None):
        super().__init__(size, is_posdef=True, is_negdef=False)

    def _scalar_multiply(self, scalar):
        return ScaledIdentityMatrix(scalar, self.shape[0])

    def _left_matrix_multiply(self, other):
        return other

    def _right_matrix_multiply(self, other):
        return other

    @property
    def eigval(self):
        return self.diagonal

    @property
    def sqrt(self):
        return self

    @property
    def eigvec(self):
        return self

    @property
    def inv(self):
        return self

    @property
    def diagonal(self):
        return np.ones(self.shape[0])

    @property
    def array(self):
        if self.shape[0] is None:
            raise RuntimeError(
                'Cannot get array representation for identity matrix with '
                'implicit size.')
        return np.identity(self.shape[0])

    @property
    def log_abs_det(self):
        return 0.

    @property
    def log_abs_det_sqrt(self):
        return 0.


class ScaledIdentityMatrix(_AbstractSymmetricMatrix):

    def __init__(self, scalar, size=None):
        if scalar == 0:
            raise ValueError('scalar must be non-zero')
        self._scalar = scalar
        super().__init__(size, is_posdef=scalar > 0, is_negdef=scalar < 0)

    def _scalar_multiply(self, scalar):
        return ScaledIdentityMatrix(scalar * self._scalar, self.shape[0])

    def _left_matrix_multiply(self, other):
        return self._scalar * other

    def _right_matrix_multiply(self, other):
        return self._scalar * other

    @property
    def eigval(self):
        return self.diagonal

    @property
    def sqrt(self):
        if not self.is_posdef:
            raise NotImplementedError(
                'Square-root supported only for positive definite '
                'matrices.')
        return ScaledIdentityMatrix(self._scalar**0.5, self.shape[0])

    @property
    def eigvec(self):
        return IdentityMatrix(self.shape[0])

    @property
    def inv(self):
        return ScaledIdentityMatrix(1 / self._scalar, self.shape[0])

    @property
    def diagonal(self):
        return self._scalar * np.ones(self.shape[0])

    @property
    def array(self):
        if self.shape[0] is None:
            raise RuntimeError(
                'Cannot get array representation for scaled identity matrix '
                'with implicit size.')
        return self._scalar * np.identity(self.shape[0])

    @property
    def log_abs_det(self):
        if self.shape[0] is None:
            raise RuntimeError(
                'Cannot get log determinant for scaled identity matrix with '
                'implicit size.')
        return self.shape[0] * np.log(abs(self._scalar))

    @property
    def grad_log_abs_det_sqrt(self):
        return 0.5 * self.shape[0] / self._scalar

    def grad_quadratic_form_inv(self, vector):
        return -np.sum(vector**2) / self._scalar**2


class PositiveScaledIdentityMatrix(ScaledIdentityMatrix):

    def __init__(self, scalar, size=None):
        if scalar < 0:
            raise ValueError('Scalar multiplier must be positive.')
        super().__init__(scalar, size)

    def _scalar_multiply(self, scalar):
        if scalar > 0:
            return PositiveScaledIdentityMatrix(
                scalar * self._scalar, self.shape[0])
        else:
            return ScaledIdentityMatrix(scalar * self._scalar, self.shape[0])

    @property
    def sqrt(self):
        return PositiveScaledIdentityMatrix(self._scalar**0.5, self.shape[0])


class DiagonalMatrix(_AbstractSymmetricMatrix):

    def __init__(self, diagonal):
        if diagonal.ndim != 1:
            raise ValueError('Specified diagonal must be a 1D array.')
        self._diagonal = diagonal
        super().__init__(diagonal.size, is_posdef=np.all(diagonal > 0),
                         is_negdef=np.all(diagonal < 0))

    @property
    def diagonal(self):
        return self._diagonal

    def _scalar_multiply(self, scalar):
        return DiagonalMatrix(self.diagonal * scalar)

    def _left_matrix_multiply(self, other):
        if other.ndim == 2:
            return self.diagonal[:, None] * other
        elif other.ndim == 1:
            return self.diagonal * other
        else:
            raise ValueError(
                'Left matrix multiplication only defined for one or two '
                'dimensional right hand sides.')

    def _right_matrix_multiply(self, other):
        return self.diagonal * other

    @property
    def eigvec(self):
        return IdentityMatrix(self.shape[0])

    @property
    def eigval(self):
        return self.diagonal

    @property
    def inv(self):
        return DiagonalMatrix(1. / self.diagonal)

    @property
    def array(self):
        return np.diag(self.diagonal)

    @property
    def grad_log_abs_det_sqrt(self):
        return 0.5 / self.diagonal

    def grad_quadratic_form_inv(self, vector):
        return -(self.inv @ vector)**2


class PositiveDiagonalMatrix(DiagonalMatrix):

    def __init__(self, diagonal):
        if not np.all(diagonal >= 0):
            raise ValueError('Diagonal values must all be positive.')
        super().__init__(diagonal)

    def _scalar_multiply(self, scalar):
        if scalar > 0:
            return PositiveDiagonalMatrix(self.diagonal * scalar)
        else:
            return DiagonalMatrix(self.diagonal * scalar)

    @property
    def sqrt(self):
        return PositiveDiagonalMatrix(self.diagonal**0.5)


class TriangularMatrix(_AbstractSquareMatrix):

    def __init__(self, array, lower=True):
        self._array = array
        self.lower = lower
        super().__init__(array.shape[0])

    def _scalar_multiply(self, scalar):
        return TriangularMatrix(self.array * scalar, self.lower)

    @property
    def inv(self):
        return InverseTriangularMatrix(self.array, lower=self.lower)

    @property
    def T(self):
        return TriangularMatrix(self.array.T, lower=not self.lower)

    @property
    def log_abs_det(self):
        return np.log(np.abs(self.diagonal)).sum()


class InverseTriangularMatrix(_AbstractSquareMatrix):

    def __init__(self, inverse_array, lower=True):
        self._inverse_array = inverse_array
        self.lower = lower
        super().__init__(inverse_array.shape[0])

    @property
    def inverse_array(self):
        return self._inverse_array

    def _scalar_multiply(self, scalar):
        return InverseTriangularMatrix(self.inverse_array / scalar, self.lower)

    def _left_matrix_multiply(self, other):
        return sla.solve_triangular(
            self.inverse_array, other, lower=self.lower)

    def _right_matrix_multiply(self, other):
        return sla.solve_triangular(
            self.inverse_array, other.T, lower=self.lower, trans=1).T

    @property
    def inv(self):
        return TriangularMatrix(self.inverse_array, lower=self.lower)

    @property
    def T(self):
        return InverseTriangularMatrix(
            self.inverse_array.T, lower=not self.lower)

    @property
    def array(self):
        return self @ np.identity(self.shape[0])

    @property
    def diagonal(self):
        return 1. / self.inverse_array.diagonal()

    @property
    def log_abs_det(self):
        return -self.inv.log_abs_det


class _AbstractTriangularFactoredDefiniteMatrix(_AbstractSymmetricMatrix):

    def __init__(self, array, factor, sign=1):
        self._array = array
        self._factor = factor
        self._sign = sign
        if array is not None:
            super().__init__(size=array.shape[0], is_posdef=(sign == 1),
                             is_negdef=(sign == -1))
        elif factor is not None:
            super().__init__(size=factor.shape[0], is_posdef=(sign == 1),
                             is_negdef=(sign == -1))
        else:
            raise RuntimeError(
                'At least one of array and factor must be specified.')

    @property
    def factor(self):
        return self._factor

    @property
    def inv(self):
        return TriangularFactoredDefiniteMatrix(
            factor=self.factor.inv.T, sign=self._sign)

    @property
    def sqrt(self):
        if self._sign == 1:
            return self.factor
        else:
            raise NotImplementedError(
                'Square-root not supported for negative definite matrices.')

    @property
    def log_abs_det(self):
        return 2 * self.factor.log_abs_det

    @property
    def log_abs_det_sqrt(self):
        return self.factor.log_abs_det


class TriangularFactoredDefiniteMatrix(
        _AbstractTriangularFactoredDefiniteMatrix):

    def __init__(self, factor, lower=True, sign=1, array=None):
        if not isinstance(factor, (TriangularMatrix, InverseTriangularMatrix)):
            factor = TriangularMatrix(factor, lower)
        super().__init__(array=array, factor=factor, sign=sign)

    def _scalar_multiply(self, scalar):
        return TriangularFactoredDefiniteMatrix(
            factor=abs(scalar)**0.5 * self.factor,
            sign=self._sign * np.sign(scalar))

    def _left_matrix_multiply(self, other):
        return self._sign * (self.factor @ (self.factor.T @ other))

    def _right_matrix_multiply(self, other):
        return self._sign * ((other @ self.factor) @ self.factor.T)

    @property
    def grad_log_abs_det_sqrt(self):
        return self.factor.inv.T.array

    def grad_quadratic_form_inv(self, vector):
        inv_factor_vector = self.factor.inv @ vector
        inv_vector = self.inv @ vector
        return -2 * self._sign * np.outer(inv_vector, inv_factor_vector)

    @property
    def array(self):
        if self._array is None:
            self._array = self._sign * (self.factor @ self.factor.array.T)
        return self._array


class DenseDefiniteMatrix(_AbstractTriangularFactoredDefiniteMatrix):

    def __init__(self, array, sign=1, factor=None):
        super().__init__(array=array, factor=factor, sign=sign)

    def _scalar_multiply(self, scalar):
        if (scalar > 0) == (self._sign == 1):
            return DenseDefiniteMatrix(
                scalar * self.array, 1,
                None if self._factor is None else
                abs(scalar)**0.5 * self._factor)
        else:
            return DenseDefiniteMatrix(
                scalar * self.array, -1,
                None if self._factor is None else
                abs(scalar)**0.5 * self._factor)

    @property
    def factor(self):
        if self._factor is None:
            self._factor = TriangularMatrix(
                nla.cholesky(self._sign * self._array), True)
        return self._factor

    @property
    def grad_log_abs_det_sqrt(self):
        return 0.5 * self.inv.array

    def grad_quadratic_form_inv(self, vector):
        inv_metric_vector = self.inv @ vector
        return -np.outer(inv_metric_vector, inv_metric_vector)


class DensePositiveDefiniteMatrix(DenseDefiniteMatrix):

    def __init__(self, array, factor=None):
        super().__init__(array=array, sign=1, factor=factor)

    def _scalar_multiply(self, scalar):
        if scalar > 0:
            return DensePositiveDefiniteMatrix(
                scalar * self.array,
                None if self._factor is None else
                abs(scalar)**0.5 * self._factor)
        else:
            return DenseNegativeDefiniteMatrix(
                scalar * self.array,
                None if self._factor is None else
                abs(scalar)**0.5 * self._factor)


class DenseNegativeDefiniteMatrix(DenseDefiniteMatrix):

    def __init__(self, array, factor=None):
        super().__init__(array=array, sign=-1, factor=factor)

    def _scalar_multiply(self, scalar):
        if scalar > 0:
            return DenseNegativeDefiniteMatrix(
                scalar * self.array,
                None if self._factor is None else
                abs(scalar)**0.5 * self._factor)
        else:
            return DensePositiveDefiniteMatrix(
                scalar * self.array,
                None if self._factor is None else
                abs(scalar)**0.5 * self._factor)


class DenseSquareMatrix(_AbstractSquareMatrix):

    def __init__(self, array, transposed=False, lu_and_piv=None):
        self._array = array
        self._transposed = transposed
        self._lu_and_piv = lu_and_piv
        super().__init__(array.shape[0])

    def _scalar_multiply(self, scalar):
        if self._lu_and_piv is None:
            return DenseSquareMatrix(scalar * self.array, self._transposed)
        else:
            old_lu, piv = self._lu_and_piv
            # Multiply upper-triangle by scalar
            new_lu = old_lu + (scalar - 1) * np.triu(old_lu)
            return DenseSquareMatrix(
                scalar * self.array, self._transposed, (new_lu, piv))

    @property
    def lu_and_piv(self):
        if self._lu_and_piv is None:
            self._lu_and_piv = sla.lu_factor(self.array)
        return self._lu_and_piv

    @property
    def log_abs_det(self):
        return np.log(np.abs(self.lu_and_piv[0].diagonal())).sum()

    @property
    def T(self):
        return DenseSquareMatrix(
            self.array.T, not self._transposed, self.lu_and_piv)

    @property
    def inv(self):
        return InverseLUFactoredSquareMatrix(
            self.array, self._transposed, self.lu_and_piv)


class InverseLUFactoredSquareMatrix(_AbstractSquareMatrix):

    def __init__(self, inverse_array, transposed, inverse_lu_and_piv):
        self._inverse_array = inverse_array
        self._inverse_lu_and_piv = inverse_lu_and_piv
        self._transposed = transposed
        super().__init__(inverse_array.shape[0])

    def _scalar_multiply(self, scalar):
        old_inv_lu, piv = self._inverse_lu_and_piv
        # Divide upper-triangle by scalar
        new_inv_lu = old_inv_lu - (scalar - 1) / scalar * np.triu(old_inv_lu)
        return InverseLUFactoredSquareMatrix(
            self._inverse_array / scalar, self._transposed, (new_inv_lu, piv))

    def _left_matrix_multiply(self, other):
        return sla.lu_solve(
            self._inverse_lu_and_piv, other, 1 if self._transposed else 0)

    def _right_matrix_multiply(self, other):
        return sla.lu_solve(
            self._inverse_lu_and_piv, other.T, 0 if self._transposed else 1).T

    @property
    def log_abs_det(self):
        return -np.log(np.abs(self._inverse_lu_and_piv[0].diagonal())).sum()

    @property
    def array(self):
        return self @ np.identity(self.shape[0])

    @property
    def inv(self):
        return DenseSquareMatrix(
            self._inverse_array, self._transposed, self._inverse_lu_and_piv)

    @property
    def T(self):
        return InverseLUFactoredSquareMatrix(
            self._inverse_array, not self._transposed,
            self._inverse_lu_and_piv)


class OrthogonalMatrix(_AbstractSquareMatrix):

    def __init__(self, array):
        self._array = array
        super().__init__(array.shape[0])

    def _scalar_multiply(self, scalar):
        return ScaledOrthogonalMatrix(scalar, self.array)

    @property
    def log_abs_det(self):
        return 0

    @property
    def T(self):
        return OrthogonalMatrix(self.array.T)

    @property
    def inv(self):
        return self.T


class ScaledOrthogonalMatrix(_AbstractSquareMatrix):

    def __init__(self, scalar, orth_array):
        self._scalar = scalar
        self._orth_array = orth_array
        super().__init__(orth_array.shape[0])

    def _scalar_multiply(self, scalar):
        return ScaledOrthogonalMatrix(scalar * self._scalar, self._orth_array)

    @property
    def array(self):
        return self._scalar * self._orth_array

    @property
    def diagonal(self):
        return self._scalar * self._orth_array.diagonal()

    @property
    def log_abs_det(self):
        return self.shape[0] * np.log(abs(self._scalar))

    @property
    def T(self):
        return ScaledOrthogonalMatrix(self._scalar, self._orth_array.T)

    @property
    def inv(self):
        return ScaledOrthogonalMatrix(1 / self._scalar, self._orth_array.T)


class EigendecomposedSymmetricMatrix(_AbstractSymmetricMatrix):

    def __init__(self, eigvec, eigval):
        if isinstance(eigvec, np.ndarray):
            eigvec = OrthogonalMatrix(eigvec)
        super().__init__(eigvec.shape[0], is_posdef=np.all(eigval > 0),
                         is_negdef=np.all(eigval < 0))
        self._eigvec = eigvec
        self._eigval = eigval
        if not isinstance(eigval, np.ndarray) or eigval.size == 1:
            self.diag_eigval = ScaledIdentityMatrix(eigval)
        else:
            self.diag_eigval = DiagonalMatrix(eigval)

    def _scalar_multiply(self, scalar):
        return EigendecomposedSymmetricMatrix(
            self.eigvec, self.eigval * scalar)

    def _left_matrix_multiply(self, other):
        return self.eigvec @ (self.diag_eigval @ (self.eigvec.T @ other))

    def _right_matrix_multiply(self, other):
        return ((other @ self.eigvec) @ self.diag_eigval) @ self.eigvec.T

    @property
    def sqrt(self):
        return EigendecomposedSymmetricMatrix(
            self.eigvec, self.eigval**0.5)

    @property
    def inv(self):
        return EigendecomposedSymmetricMatrix(
            self.eigvec, 1 / self.eigval)

    @property
    def array(self):
        if self.shape[0] is None:
            raise RuntimeError(
                'Cannot get array representation for symmetric '
                'eigendecomposed matrix with implicit size.')
        return self @ np.identity(self.shape[0])


class SoftAbsRegularisedPositiveDefiniteMatrix(EigendecomposedSymmetricMatrix):

    def __init__(self, symmetric_array, softabs_coeff):
        """
        Args:
            softabs_coeff (float): Positive regularisation coefficient for
                smooth approximation to absolute value. As the value tends to
                infinity the approximation becomes increasingly close to the
                absolute function.
        """
        if softabs_coeff <= 0:
            raise ValueError('softabs_coeff must be positive.')
        self._softabs_coeff = softabs_coeff
        self.unreg_eigval, eigvec = nla.eigh(symmetric_array)
        eigval = self.softabs(self.unreg_eigval)
        super().__init__(eigvec, eigval)

    def softabs(self, x):
        """Smooth approximation to absolute function."""
        return x / np.tanh(x * self._softabs_coeff)

    def grad_softabs(self, x):
        """Derivative of smooth approximation to absolute function."""
        return (
            1. / np.tanh(self._softabs_coeff * x) -
            self._softabs_coeff * x / np.sinh(self._softabs_coeff * x)**2)

    @property
    def grad_log_abs_det_sqrt(self):
        return 0.5 * (
            self.eigvec @ DiagonalMatrix(
                self.grad_softabs(self.unreg_eigval) / self.eigval) @
            self.eigvec.T)

    def grad_quadratic_form_inv(self, vector):
        num_j_mtx = self.eigval[:, None] - self.eigval[None, :]
        num_j_mtx += np.diag(self.grad_softabs(self.unreg_eigval))
        den_j_mtx = self.unreg_eigval[:, None] - self.unreg_eigval[None, :]
        np.fill_diagonal(den_j_mtx, 1)
        j_mtx = num_j_mtx / den_j_mtx
        e_vct = (self.eigvec.T @ vector) / self.eigval
        return -(
            (self.eigvec @ (np.outer(e_vct, e_vct) * j_mtx)) @ self.eigvec.T)


class MatrixProduct(_AbstractMatrix):

    def __init__(self, matrices):
        self._matrices = matrices
        super().__init__((matrices[0].shape[0], matrices[-1].shape[1]))

    def _scalar_multiply(self, scalar):
        return MatrixProduct((
            ScaledIdentityMatrix(scalar, self.shape[0]), *self.matrices))

    def _left_matrix_multiply(self, other):
        for matrix in reversed(self._matrices):
            other = matrix @ other
        return other

    def _right_matrix_multiply(self, other):
        for matrix in self._matrices:
            other = other @ matrix
        return other

    @property
    def T(self):
        return MatrixProduct(
            [matrix.T for matrix in reversed(self._matrices)])

    @property
    def array(self):
        return self @ np.identity(self.shape[1])



"""Hamiltonian systems encapsulating energy functions and their derivatives."""

import logging
import numpy as np
from mici.states import cache_in_state, multi_cache_in_state
from mici.matrices import (
    IdentityMatrix, PositiveScaledIdentityMatrix, PositiveDiagonalMatrix,
    DenseSquareMatrix, TriangularFactoredDefiniteMatrix, DenseDefiniteMatrix,
    DensePositiveDefiniteMatrix, EigendecomposedSymmetricMatrix,
    SoftAbsRegularisedPositiveDefiniteMatrix)
from mici.autodiff import autodiff_fallback


class _HamiltonianSystem(object):
    """Base class for Hamiltonian systems.

    The Hamiltonian function `h` is assumed to have the general form

        h(pos, mom) = h1(pos) + h2(pos, mom)

    where `pos` and `mom` are the position and momentum variables respectively,
    and `h1(pos)` and `h2(pos, mom)` Hamiltonian components. The exact
    Hamiltonian flow for the `h1` component can be computed however depending
    on the form of `h2` the corresponding Hamiltonian flow may or may not be
    simulable.

    By default `h1` is assumed to correspond to the negative logarithm of an
    unnormalised density on the position variables with respect to the Lebesgue
    measure, with the corresponding distribution on the position space being
    the target distribution it is wished to draw approximate samples from.
    """

    def __init__(self, neg_log_dens, grad_neg_log_dens=None):
        """
        Args:
            neg_log_dens (callable): Function which given a position vector
                returns the negative logarithm of an unnormalised probability
                density on the position space with respect to the Lebesgue
                measure, with the corresponding distribution on the position
                space being the target distribution it is wished to draw
                approximate samples from.
            grad_neg_log_dens (callable or None): Function which given a
                position vector returns the derivative of the negative
                logarithm of the unnormalised density specified by
                `neg_log_dens` with respect to the position vector argument.
                Optionally the function may instead return a pair of values
                with the first being the value of the `neg_log_dens` evaluated
                at the passed position vector and the second being the value of
                its derivative with respect to the position argument. If `None`
                is passed (the default) an automatic differentiation fallback
                will be used to attempt to construct the derivative of
                `neg_log_dens` automatically.

        """
        self._neg_log_dens = neg_log_dens
        self._grad_neg_log_dens = autodiff_fallback(
            grad_neg_log_dens, neg_log_dens,
            'grad_and_value', 'grad_neg_log_dens')

    @cache_in_state('pos')
    def neg_log_dens(self, state):
        return self._neg_log_dens(state.pos)

    @multi_cache_in_state(['pos'], ['grad_neg_log_dens', 'neg_log_dens'])
    def grad_neg_log_dens(self, state):
        return self._grad_neg_log_dens(state.pos)

    def h1(self, state):
        return self.neg_log_dens(state)

    def dh1_dpos(self, state):
        return self.grad_neg_log_dens(state)

    def h1_flow(self, state, dt):
        state.mom -= dt * self.dh1_dpos(state)

    def h2(self, state):
        raise NotImplementedError()

    def h(self, state):
        return self.h1(state) + self.h2(state)

    def dh_dpos(self, state):
        if hasattr(self, 'dh2_dpos'):
            return self.dh1_dpos(state) + self.dh2_dpos(state)
        else:
            return self.dh1_dpos(state)

    def dh_dmom(self, state):
        return self.dh2_dmom(state)

    def sample_momentum(self, state, rng):
        raise NotImplementedError()


class EuclideanMetricSystem(_HamiltonianSystem):
    """Hamiltonian system with fixed covariance Gaussian-distributed momenta.

    The momentum variables are taken to be independent of the position
    variables and with a zero-mean Gaussian marginal distribution with
    covariance specified by a fixed positive-definite matrix (metric tensor),
    so that the `h2` Hamiltonian component is

        h2(pos, mom) = 0.5 * mom @ inv(metric) @ mom

    where `pos` and `mom` are the position and momentum variables respectively,
    and `inv(metric)` is the matrix inverse of the metric tensor.
    """

    def __init__(self, neg_log_dens, metric=None, grad_neg_log_dens=None):
        """
        Args:
            neg_log_dens (callable): Function which given a position vector
                returns the negative logarithm of an unnormalised probability
                density on the position space with respect to the Lebesgue
                measure, with the corresponding distribution on the position
                space being the target distribution it is wished to draw
                approximate samples from.
            metric (None or array or Matrix): Matrix object corresponding to
                matrix representation of metric on position space and
                covariance of Gaussian marginal distribution on momentum
                vector. If `None` is passed (the default), the identity matrix
                will be used. If a 1D array is passed then this is assumed to
                specify a metric with diagonal matrix representation and the
                array to the matrix diagonal. If a 2D array is passed then this
                is assumed to specify a metric with a dense positive definite
                matrix representation specified by the array.
            grad_neg_log_dens (callable or None): Function which given a
                position vector returns the derivative of the negative
                logarithm of the unnormalised density specified by
                `neg_log_dens` with respect to the position vector argument.
                Optionally the function may instead return a pair of values
                with the first being the value of the `neg_log_dens` evaluated
                at the passed position vector and the second being the value of
                its derivative with respect to the position argument. If `None`
                is passed (the default) it will be attempted to use automatic
                differentiation to construct the derivative of `neg_log_dens`
                automatically.
        """
        super().__init__(neg_log_dens, grad_neg_log_dens)
        if metric is None:
            self.metric = IdentityMatrix()
        elif isinstance(metric, np.ndarray):
            if metric.ndim == 1:
                self.metric = PositiveDiagonalMatrix(metric)
            elif metric.ndim == 2:
                self.metric = DensePositiveDefiniteMatrix(metric)
            else:
                raise ValueError('If NumPy ndarray value is used for `metric`'
                                 ' must be either 1D (diagonal matrix) or 2D '
                                 '(dense positive definite matrix)')
        else:
            self.metric = metric

    @cache_in_state('mom')
    def h2(self, state):
        return 0.5 * state.mom @ self.metric.inv @ state.mom

    @cache_in_state('mom')
    def dh2_dmom(self, state):
        return self.metric.inv @ state.mom

    def h2_flow(self, state, dt):
        state.pos += dt * self.dh2_dmom(state)

    def dh2_flow_dmom(self, dt):
        return dt * self.metric.inv, IdentityMatrix(self.metric.shape[0])

    def sample_momentum(self, state, rng):
        return self.metric.sqrt @ rng.standard_normal(state.pos.shape)


class GaussianEuclideanMetricSystem(EuclideanMetricSystem):
    """Euclidean Hamiltonian system with a tractable Gaussian component.

    The momentum variables are taken to be independent of the position
    variables and with a zero-mean Gaussian marginal distribution with
    covariance specified by a fixed positive-definite matrix (metric tensor).

    Additionally the target distribution on the position variables is assumed
    to be defined by an unnormalised density with respect to the standard
    Gaussian measure on the position space (with identity covariance and zero
    mean), with the Hamiltonian component `h1` corresponding to the negative
    logarithm of this density rather than the density with respect to the
    Lebesgue measure on the position space. The Hamiltonian component function
    `h2` is therefore assumed to have the form

         h2(pos, mom) = 0.5 * pos @ pos + 0.5 * mom @ inv(metric) @ mom

    where `pos` and `mom` are the position and momentum variables respectively,
    and `inv(metric)` is the matrix inverse of the metric tensor. In this case
    the Hamiltonian flow due to the quadratic `h2` component can be solved for
    analytically, allowing an integrator to be defined using this alternative
    splitting of the Hamiltonian [1].

    References:

      1. Shahbaba, B., Lan, S., Johnson, W.O. and Neal, R.M., 2014. Split
         Hamiltonian Monte Carlo. Statistics and Computing, 24(3), pp.339-349.
    """

    def __init__(self, neg_log_dens, metric=None, grad_neg_log_dens=None):
        """
        Args:
            neg_log_dens (callable): Function which given a position vector
                returns the negative logarithm of an unnormalised probability
                density on the position space with respect to the standard
                Gaussian measure on the position space, with the corresponding
                distribution on the position space being the target
                distribution it is wished to draw approximate samples from.
            metric (Matrix or None): Matrix object corresponding to covariance
                of Gaussian marginal distribution on momentum vector. If `None`
                is passed (the default), the identity matrix will be used.
            grad_neg_log_dens (callable or None): Function which given a
                position vector returns the derivative of the negative
                logarithm of the unnormalised density specified by
                `neg_log_dens` with respect to its position vector argument.
                Optionally the function may instead return a pair of values
                with the first being the value of the `neg_log_dens` evaluated
                at the passed position vector and the second being the value of
                its derivative with respect to the position argument. If `None`
                is passed (the default) an automatic differentiation backend
                will be used to construct the derivative of `neg_log_dens`
                automatically if available.
        """
        super().__init__(neg_log_dens, metric, grad_neg_log_dens)

    def h2(self, state):
        return (0.5 * state.pos @ state.pos +
                0.5 * state.mom @ self.metric.inv @ state.mom)

    @cache_in_state('mom')
    def dh2_dmom(self, state):
        return self.metric.inv @ state.mom

    @cache_in_state('mom')
    def dh2_dpos(self, state):
        return state.pos

    def h2_flow(self, state, dt):
        omega = 1. / self.metric.eigval**0.5
        sin_omega_dt, cos_omega_dt = np.sin(omega * dt), np.cos(omega * dt)
        eigvec_T_pos = self.metric.eigvec.T @ state.pos
        eigvec_T_mom = self.metric.eigvec.T @ state.mom
        state.pos = self.metric.eigvec @ (
            cos_omega_dt * eigvec_T_pos +
            (sin_omega_dt * omega) * eigvec_T_mom)
        state.mom = self.metric.eigvec @ (
            cos_omega_dt * eigvec_T_mom -
            (sin_omega_dt / omega) * eigvec_T_pos)

    def dh2_flow_dmom(self, dt):
        omega = 1. / self.metric.eigval**0.5
        sin_omega_dt, cos_omega_dt = np.sin(omega * dt), np.cos(omega * dt)
        return (
            EigendecomposedSymmetricMatrix(
                self.metric.eigvec, sin_omega_dt * omega),
            EigendecomposedSymmetricMatrix(
                self.metric.eigvec, cos_omega_dt))


class _ConstrainedEuclideanMetricSystem(EuclideanMetricSystem):
    """Base class for Euclidean Hamiltonian systems subject to constraints.

    The system is assumed to be subject to a set of holonomic constraints on
    the position component of the state. These constraints are specified by a
    vector constraint function which takes as argument the position component,
    and which is equal to zero in all components when the constraints are
    satisfied. The constraint function implicitly defines a manifold embedded
    in the position space of constraint satisfying configurations. There are
    also implicitly a set of constraints on the momentum component of the state
    due to the requirment that velocity (momentum pre-multiplied by inverse
    metric) is always tangential to the constraint manifold.
    """

    def __init__(self, neg_log_dens, constr, metric=None,
                 dens_wrt_hausdorff=True, grad_neg_log_dens=None,
                 jacob_constr=None):
        super().__init__(neg_log_dens=neg_log_dens, metric=metric,
                         grad_neg_log_dens=grad_neg_log_dens)
        self._constr = constr
        self.dens_wrt_hausdorff = dens_wrt_hausdorff
        self._jacob_constr = autodiff_fallback(
            jacob_constr, constr, 'jacobian_and_value', 'jacob_constr')

    @cache_in_state('pos')
    def constr(self, state):
        return self._constr(state.pos)

    @multi_cache_in_state(['pos'], ['jacob_constr', 'constr'])
    def jacob_constr(self, state):
        return self._jacob_constr(state.pos)

    def jacob_constr_inner_product(
            self, jacob_constr_1, inner_product_matrix, jacob_constr_2=None):
        raise NotImplementedError()

    @cache_in_state('pos')
    def gram(self, state):
        return self.jacob_constr_inner_product(
            self.jacob_constr(state), self.metric.inv)

    def inv_gram(self, state):
        return self.gram(state).inv

    def log_det_sqrt_gram(self, state):
        return self.gram(state).log_abs_det_sqrt

    def grad_log_det_sqrt_gram(self, state):
        raise NotImplementedError()

    def h1(self, state):
        if self.dens_wrt_hausdorff:
            return self.neg_log_dens(state)
        else:
            return self.neg_log_dens(state) + self.log_det_sqrt_gram(state)

    def dh1_dpos(self, state):
        if self.dens_wrt_hausdorff:
            return self.grad_neg_log_dens(state)
        else:
            return (self.grad_neg_log_dens(state) +
                    self.grad_log_det_sqrt_gram(state))

    def project_onto_cotangent_space(self, mom, state):
        jacob_constr = self.jacob_constr(state)
        # Use parenthesis to force right-to-left evaluation to avoid
        # matrix-matrix products
        mom -= (self.jacob_constr(state).T @ (
                    self.inv_gram(state) @ (
                        self.jacob_constr(state) @ (self.metric.inv @ mom))))

    def sample_momentum(self, state, rng):
        mom = super().sample_momentum(state, rng)
        self.project_onto_cotangent_space(mom, state)
        return mom


class DenseConstrainedEuclideanMetricSystem(_ConstrainedEuclideanMetricSystem):
    """Euclidean Hamiltonian systems subject to a dense set of constraints.

    The system is assumed to be subject to a set of holonomic constraints on
    the position component of the state. These constraints are specified by a
    vector constraint function which takes as argument the position component,
    and which is equal to zero in all components when the constraints are
    satisfied. The constraint function implicitly defines a manifold embedded
    in the position space of constraint satisfying configurations. There are
    also implicitly a set of constraints on the momentum component of the state
    due to the requirment that velocity (momentum pre-multiplied by inverse
    metric) is always tangential to the constraint manifold.
    """

    def __init__(self, neg_log_dens, constr, metric=None,
                 dens_wrt_hausdorff=True, grad_neg_log_dens=None,
                 jacob_constr=None, mhp_constr=None):
        super().__init__(neg_log_dens=neg_log_dens, constr=constr,
                         metric=metric, dens_wrt_hausdorff=dens_wrt_hausdorff,
                         grad_neg_log_dens=grad_neg_log_dens,
                         jacob_constr=jacob_constr)
        if not dens_wrt_hausdorff:
            self._mhp_constr = autodiff_fallback(
                mhp_constr, constr, 'mhp_jacobian_and_value', 'mhp_constr')

    @multi_cache_in_state(
        ['pos'], ['mhp_constr', 'jacob_constr', 'constr'])
    def mhp_constr(self, state):
        return self._mhp_constr(state.pos)

    def jacob_constr_inner_product(
            self, jacob_constr_1, inner_product_matrix, jacob_constr_2=None):
        if jacob_constr_2 is None or jacob_constr_2 is jacob_constr_1:
            return DenseDefiniteMatrix(
                jacob_constr_1 @ inner_product_matrix @ jacob_constr_1.T,
                sign=1 if inner_product_matrix.is_posdef else -1)
        else:
            return DenseSquareMatrix(
                jacob_constr_1 @ inner_product_matrix @ jacob_constr_2.T)

    @cache_in_state('pos')
    def grad_log_det_sqrt_gram(self, state):
        return self.mhp_constr(state)(
            self.inv_gram(state) @ self.jacob_constr(state) @ self.metric.inv)


class GaussianDenseConstrainedEuclideanMetricSystem(
        GaussianEuclideanMetricSystem, DenseConstrainedEuclideanMetricSystem):

    def __init__(self, neg_log_dens, constr, metric=None,
                 dens_wrt_hausdorff=True, grad_neg_log_dens=None,
                 jacob_constr=None, mhp_constr=None):
        DenseConstrainedEuclideanMetricSystem.__init__(
            self, neg_log_dens=neg_log_dens, constr=constr, metric=metric,
            dens_wrt_hausdorff=dens_wrt_hausdorff,
            grad_neg_log_dens=grad_neg_log_dens, jacob_constr=jacob_constr,
            mhp_constr=mhp_constr)


class RiemannianMetricSystem(_HamiltonianSystem):
    """Riemannian Hamiltonian system with a position dependent metric tensor.

    The momentum variables are assumed to have a zero-mean Gaussian conditional
    distribution given the position variables, with covariance specified by a
    position dependent positive-definite metric tensor [1]. Due to the coupling
    between the position and momentum variables in the quadratic form of the
    negative log density of the Gaussian conditional distribution on the
    momentum variables, the Hamiltonian system is non-separable, requiring use
    of a numerical integrator with implicit steps.

    References:

      1. Girolami, M. and Calderhead, B., 2011. Riemann manifold Langevin and
         Hamiltonian Monte Varlo methods. Journal of the Royal Statistical
         Society: Series B (Statistical Methodology), 73(2), pp.123-214.
    """

    def __init__(self, neg_log_dens, metric_matrix_class, metric_func,
                 vjp_metric_func=None, grad_neg_log_dens=None,
                 metric_kwargs=None):
        self._metric_matrix_class = metric_matrix_class
        self._metric_func = metric_func
        self._vjp_metric_func = autodiff_fallback(
            vjp_metric_func, metric_func, 'vjp_and_value', 'vjp_metric_func')
        self._metric_kwargs = {} if metric_kwargs is None else metric_kwargs
        super().__init__(neg_log_dens, grad_neg_log_dens)

    @cache_in_state('pos')
    def metric_func(self, state):
        return self._metric_func(state.pos)

    @multi_cache_in_state(['pos'], ['vjp_metric_func', 'metric_func'])
    def vjp_metric_func(self, state):
        return self._vjp_metric_func(state.pos)

    @cache_in_state('pos')
    def metric(self, state):
        return self._metric_matrix_class(
            self.metric_func(state), **self._metric_kwargs)

    def h(self, state):
        return self.h1(state) + self.h2(state)

    def h1(self, state):
        return self.neg_log_dens(state) + self.metric(state).log_abs_det_sqrt

    def dh1_dpos(self, state):
        # Evaluate VJP of metric function before metric as metric value will
        # be computed in forward pass and cached
        vjp_metric = self.vjp_metric_func(state)
        return (self.grad_neg_log_dens(state) +
                vjp_metric(self.metric(state).grad_log_abs_det_sqrt))

    def h2(self, state):
        return 0.5 * state.mom @ self.metric(state).inv @ state.mom

    def dh2_dpos(self, state):
        # Evaluate VJP of metric function before metric as metric value will
        # be computed in forward pass and cached
        vjp_metric = self.vjp_metric_func(state)
        return 0.5 * vjp_metric(
            self.metric(state).grad_quadratic_form_inv(state.mom))

    def dh2_dmom(self, state):
        return self.metric(state).inv @ state.mom

    def sample_momentum(self, state, rng):
        return self.metric(state).sqrt @ rng.normal(size=state.pos.shape)


class ScalarRiemannianMetricSystem(RiemannianMetricSystem):

    def __init__(self, neg_log_dens, metric_scalar_func,
                 vjp_metric_scalar_func=None, grad_neg_log_dens=None):
        super().__init__(
            neg_log_dens, PositiveScaledIdentityMatrix, metric_scalar_func,
            vjp_metric_scalar_func, grad_neg_log_dens)


class DiagonalRiemannianMetricSystem(RiemannianMetricSystem):

    def __init__(self, neg_log_dens, metric_diagonal_func,
                 vjp_metric_diagonal_func=None, grad_neg_log_dens=None):
        super().__init__(
            neg_log_dens, PositiveDiagonalMatrix, metric_diagonal_func,
            vjp_metric_diagonal_func, grad_neg_log_dens)


class CholeskyFactoredRiemannianMetricSystem(RiemannianMetricSystem):

    def __init__(self, neg_log_dens, metric_chol_func,
                 vjp_metric_chol_func=None, grad_neg_log_dens=None):
        super().__init__(
            neg_log_dens, TriangularFactoredDefiniteMatrix,
            metric_chol_func, vjp_metric_chol_func, grad_neg_log_dens)


class DenseRiemannianMetricSystem(RiemannianMetricSystem):

    def __init__(self, neg_log_dens, metric_func,
                 vjp_metric_func=None, grad_neg_log_dens=None):
        super().__init__(
            neg_log_dens, DensePositiveDefiniteMatrix, metric_func,
            vjp_metric_func, grad_neg_log_dens)


class SoftAbsRiemannianMetricSystem(RiemannianMetricSystem):
    """SoftAbs Riemmanian metric Hamiltonian system.

    Hamiltonian system with a position dependent metric tensor which is
    specified to be an eigenvalue-regularised transformation of the Hessian of
    the potential energy function (with the 'soft-absolute' regularisation
    ensuring all the eigenvalues are strictly positive and so the resulting
    metric tensor is positive definite everywhere).

    References:

    1. Betancourt, M., 2013. A general metric for Riemannian manifold
       Hamiltonian Monte Carlo. In Geometric science of information
       (pp. 327-334).
    """

    def __init__(self, neg_log_dens, grad_neg_log_dens=None,
                 hess_neg_log_dens=None, mtp_neg_log_dens=None,
                 softabs_coeff=1.):
        self._hess_neg_log_dens = autodiff_fallback(
            hess_neg_log_dens, neg_log_dens, 'hessian_grad_and_value',
            'neg_log_dens')
        self._mtp_neg_log_dens = autodiff_fallback(
            mtp_neg_log_dens, neg_log_dens, 'mtp_hessian_grad_and_value',
            'mtp_neg_log_dens')
        super().__init__(neg_log_dens,
                         SoftAbsRegularisedPositiveDefiniteMatrix,
                         self._hess_neg_log_dens, self._mtp_neg_log_dens,
                         grad_neg_log_dens,
                         metric_kwargs={'softabs_coeff': softabs_coeff})

    def metric_func(self, state):
        return self.hess_neg_log_dens(state)

    def vjp_metric_func(self, state):
        return self.mtp_neg_log_dens(state)

    @multi_cache_in_state(
        ['pos'], ['hess_neg_log_dens', 'grad_neg_log_dens', 'neg_log_dens'])
    def hess_neg_log_dens(self, state):
        return self._hess_neg_log_dens(state.pos)

    @multi_cache_in_state(
        ['pos'], ['mtp_neg_log_dens', 'hess_neg_log_dens',
                  'grad_neg_log_dens', 'neg_log_dens'])
    def mtp_neg_log_dens(self, state):
        return self._mtp_neg_log_dens(state.pos)



"""Exception types."""


class IntegratorError(RuntimeError):
    """Error raised when integrator step fails."""


class NonReversibleStepError(IntegratorError):
    """Error raised when integrator step fails reversibility check."""


class ConvergenceError(IntegratorError):
    """Error raised when solver fails to converge within allowed iterations."""



"""Monte Carlo sampler classes for peforming inference."""

import os
import logging
import tempfile
import signal
from collections import OrderedDict
import numpy as np
import mici
import mici.transitions as trans
from mici.states import ChainState
from mici.utils import get_size, get_valid_filename

try:
    import tqdm
    import tqdm.auto as tqdm_auto
    TQDM_AVAILABLE = True
except ImportError:
    TQDM_AVAILABLE = False
try:
    import randomgen
    RANDOMGEN_AVAILABLE = True
except ImportError:
    RANDOMGEN_AVAILABLE = False
# Preferentially import Pool from multiprocess library if available as able
# to serialise much wider range of types including autograd functions
try:
    from multiprocess import Pool
except ImportError:
    from multiprocessing import Pool


logger = logging.getLogger(__name__)


def _ignore_sigint_initialiser():
    """Initialiser for multi-process workers to force ignoring SIGINT."""
    signal.signal(signal.SIGINT, signal.SIG_IGN)


class MarkovChainMonteCarloMethod(object):
    """Generic Markov chain Monte Carlo (MCMC) sampler.

    Generates a Markov chain from some initial state by iteratively applying
    a sequence of Markov transition operators.
    """

    def __init__(self, rng, transitions):
        """
        Args:
            rng: Numpy RandomState random number generator instance.
            transitions: Ordered dictionary of Markov chain transitions to
                sequentially sample from on each chain iteration.
        """
        self.rng = rng
        self.transitions = transitions

    def _generate_memmap_filename(self, dir_path, prefix, key, index):
        key_str = get_valid_filename(str(key))
        if index is None:
            index = 0
        return os.path.join(dir_path, f'{prefix}_{index}_{key_str}.npy')

    def _open_new_memmap(self, filename, shape, dtype, default_val):
        memmap = np.lib.format.open_memmap(
            filename, dtype=dtype, mode='w+', shape=shape)
        memmap[:] = default_val
        return memmap

    def _memmaps_to_filenames(self, obj):
        if isinstance(obj, np.memmap):
            return obj.filename
        elif isinstance(obj, dict):
            return {k: self._memmaps_to_filenames(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._memmaps_to_filenames(v) for v in obj]

    def _init_chain_stats(self, n_sample, memmap_enabled, memmap_path,
                          chain_index):
        chain_stats = {}
        for trans_key, trans in self.transitions.items():
            chain_stats[trans_key] = {}
            if hasattr(trans, 'statistic_types'):
                for key, (dtype, val) in trans.statistic_types.items():
                    if memmap_enabled:
                        filename = self._generate_memmap_filename(
                            memmap_path, 'stats', f'{trans_key}_{key}',
                            chain_index)
                        chain_stats[trans_key][key] = self._open_new_memmap(
                            filename, (n_sample,), dtype, val)
                    else:
                        chain_stats[trans_key][key] = np.full(
                            n_sample, val, dtype)
        return chain_stats

    def _sample_chain(self, rng, n_sample, init_state, trace_funcs,
                      chain_index, parallel_chains, memmap_enabled=False,
                      memmap_path=None, monitor_stats=None):
        for trans_key, transition in self.transitions.items():
            for var_key in transition.state_variables:
                if var_key not in init_state:
                    raise ValueError(
                        f'init_state does contain have {var_key} value '
                        f'required by {trans_key} transition.')
        if not isinstance(init_state, (ChainState, dict)):
            raise TypeError(
                'init_state should be a dictionary or `ChainState`.')
        state = (ChainState(**init_state) if isinstance(init_state, dict)
                 else init_state)
        chain_stats = self._init_chain_stats(
            n_sample, memmap_enabled, memmap_path, chain_index)
        # Initialise chain trace arrays
        traces = {}
        for trace_func in trace_funcs:
            for key, val in trace_func(state).items():
                val = np.array(val) if np.isscalar(val) else val
                init = np.nan if np.issubdtype(val.dtype, np.inexact) else 0
                if memmap_enabled:
                    filename = self._generate_memmap_filename(
                        memmap_path, 'trace', key, chain_index)
                    traces[key] = self._open_new_memmap(
                        filename, (n_sample,) + val.shape, val.dtype, init)
                else:
                    traces[key] = np.full(
                        (n_sample,) + val.shape, init, val.dtype)
        total_return_nbytes = get_size(chain_stats) + get_size(traces)
        # Check if running in parallel and if total number of bytes to be
        # returned exceeds pickle limit
        if parallel_chains and total_return_nbytes > 2**31 - 1:
            raise RuntimeError(
                f'Total number of bytes allocated for arrays to be returned '
                f'({total_return_nbytes / 2**30:.2f} GiB) exceeds size limit '
                f'for returning results of a process (2 GiB). Try rerunning '
                f'with chain memory-mapping enabled (`memmap_enabled=True`).')
        if TQDM_AVAILABLE:
            kwargs = {
                'desc': f'Chain {0 if chain_index is None else chain_index}',
                'unit': 'it',
                'dynamic_ncols': True,
            }
            if parallel_chains:
                sample_range = tqdm_auto.trange(
                    n_sample, **kwargs, position=chain_index)
            else:
                sample_range = tqdm.trange(n_sample, **kwargs)
        else:
            sample_range = range(n_sample)
        try:
            for sample_index in sample_range:
                for trans_key, transition in self.transitions.items():
                    state, trans_stats = transition.sample(state, rng)
                    if trans_stats is not None:
                        if trans_key not in chain_stats:
                            logger.warning(
                                f'Transition {trans_key} returned statistics '
                                f'but has no `statistic_types` attribute.')
                        for key, val in trans_stats.items():
                            if key in chain_stats[trans_key]:
                                chain_stats[trans_key][key][sample_index] = val
                for trace_func in trace_funcs:
                    for key, val in trace_func(state).items():
                        traces[key][sample_index] = val
                if TQDM_AVAILABLE and monitor_stats is not None:
                    postfix_stats = {}
                    for (trans_key, stats_key) in monitor_stats:
                        if (trans_key not in chain_stats or
                                stats_key not in chain_stats[trans_key]):
                            logger.warning(
                                f'Statistics key pair {(trans_key, stats_key)}'
                                f' to be monitored is not valid.')
                        print_key = f'mean({stats_key})'
                        postfix_stats[print_key] = np.mean(
                            chain_stats[trans_key][stats_key][:sample_index+1])
                    sample_range.set_postfix(postfix_stats)
        except KeyboardInterrupt:
            if memmap_enabled:
                for trace in traces.values():
                    trace.flush()
                for trans_stats in chain_stats.values():
                    for stat in trans_stats.values():
                        stat.flush()
        else:
            # If not interrupted increment sample_index so that it equals
            # n_sample to flag chain completed sampling
            sample_index += 1
        if parallel_chains and memmap_enabled:
                trace_filenames = self._memmaps_to_filenames(traces)
                stats_filenames = self._memmaps_to_filenames(chain_stats)
                return trace_filenames, stats_filenames, sample_index
        return state, traces, chain_stats, sample_index

    def __preprocess_kwargs(self, kwargs):
        if 'memmap_enabled' not in kwargs:
            kwargs['memmap_enabled'] = False
        # Create temporary directory if memory mapping and no path provided
        if kwargs['memmap_enabled'] and 'memmap_path' not in kwargs:
            kwargs['memmap_path'] = tempfile.mkdtemp()
        return kwargs

    def sample_chain(self, n_sample, init_state, trace_funcs, **kwargs):
        """Sample a Markov chain from a given initial state.

        Performs a specified number of chain iterations (each of which may be
        composed of multiple individual Markov transitions), recording the
        outputs of functions of the sampled chain state after each iteration.

        Args:
            n_sample (int): Number of samples (iterations) to draw per chain.
            init_state (ChainState or dict): Initial chain state. Either a
                `ChainState` object or a dictionary with entries specifying
                initial values for all state variables used by chain
                transition `sample` methods.
            trace_funcs (Iterable[callable]): List of functions which compute
                the variables to be recorded at each chain iteration, with each
                trace function being passed the current state and returning a
                dictionary of scalar or array values corresponding to the
                variable(s) to be stored. The keys in the returned dictionaries
                are used to index the trace arrays in the returned traces
                dictionary. If a key appears in multiple dictionaries only the
                the value corresponding to the last trace function to return
                that key will be stored.

        Kwargs:
            memmap_enabled (bool): Whether to memory-map arrays used to store
                chain data to files on disk to avoid excessive system memory
                usage for long chains and/or large chain states. The chain data
                is written to `.npy` files in the directory specified by
                `memmap_path` (or a temporary directory if not provided). These
                files persist after the termination of the function so should
                be manually deleted when no longer required. Default is to
                for memory mapping to be disabled.
            memmap_path (str): Path to directory to write memory-mapped chain
                data to. If not provided, a temporary directory will be created
                and the chain data written to files there.
            monitor_stats (Iterable[tuple(str, str)]): List of tuples of string
                key pairs, with first entry the key of a Markov transition in
                the `transitions` dict passed to the the `__init__` method and
                the second entry the key of a chain statistic that will be
                returned in the `chain_stats` dictionary. The mean over samples
                computed so far of the chain statistics associated with any
                valid key-pairs will be monitored during sampling  by printing
                as postfix to progress bar (if `tqdm` is installed).

        Returns:
            final_state (ChainState): State of chain after final iteration. May
                be used to resume sampling a chain by passing as the initial
                state to a new `sample_chain` call.
            traces (dict[str, array]): Dictionary of chain trace arrays. Values
                in dictionary are arrays of variables outputted by trace
                functions in `trace_funcs` with leading dimension of array
                corresponding to the sampling (draw) index. The key for each
                value is the corresponding key in the dictionary returned by
                the trace function which computed the traced value.
            chain_stats (dict[str, dict[str, array]]): Dictionary of chain
                transition statistic dictionaries. Values in outer dictionary
                are dictionaries of statistics for each chain transition, keyed
                by the string key for the transition. The values in each inner
                transition dictionary are arrays of chain statistic values with
                the leading dimension of each array corresponding to the
                sampling (draw) index. The key for each value is a string
                description of the corresponding integration transition
                statistic.
        """
        kwargs = self.__preprocess_kwargs(kwargs)
        final_state, traces, chain_stats, sample_index = self._sample_chain(
            rng=self.rng, n_sample=n_sample, init_state=init_state,
            trace_funcs=trace_funcs, chain_index=None, parallel_chains=False,
            **kwargs)
        if sample_index != n_sample:
            # Sampling interrupted therefore truncate returned arrays
            # Using resize methods makes agnostic to whether array is
            # memory mapped or not
            for trans_stats in chain_stats.values():
                for key in trans_stats:
                    trans_stats[key].resize(
                        (sample_index,) + trans_stats[key].shape[1:])
            for key in traces:
                traces[key].resize(
                    (sample_index,) + traces[key].shape[1:])
            logger.error(
                f'Sampling manually interrupted at iteration {sample_index}. '
                f'Arrays containing chain traces and statistics computed '
                f'before interruption will be returned.')
        return final_state, traces, chain_stats

    def _collate_chain_outputs(self, n_sample, chain_outputs, load_memmaps):
        final_states_stack = []
        traces_stack = {}
        n_chain = len(chain_outputs)
        chain_stats_stack = {}
        for chain_index, chain_output in enumerate(chain_outputs):
            final_state, traces, chain_stats, sample_index = chain_output
            final_states_stack.append(final_state)
            for key, val in traces.items():
                if load_memmaps:
                    val = np.lib.format.open_memmap(val)
                if sample_index != n_sample:
                    # Sampling interrupted therefore truncate returned arrays
                    # Using resize methods makes agnostic to whether array is
                    # memory mapped or not
                    val.resize((sample_index,) + val.shape[1:], refcheck=False)
                if chain_index == 0:
                    traces_stack[key] = [val]
                else:
                    traces_stack[key].append(val)
            for trans_key, trans_stats in chain_stats.items():
                if chain_index == 0:
                    chain_stats_stack[trans_key] = {}
                for key, val in trans_stats.items():
                    if load_memmaps:
                        val = np.lib.format.open_memmap(val)
                    if sample_index != n_sample:
                        # Sampling interrupted therefore truncate returned
                        # arrays Using resize methods makes agnostic to whether
                        # array is memory mapped or not
                        val.resize((sample_index,) + val.shape[1:],
                                   refcheck=False)
                    if chain_index == 0:
                        chain_stats_stack[trans_key][key] = [val]
                    else:
                        chain_stats_stack[trans_key][key].append(val)
        return final_states_stack, traces_stack, chain_stats_stack

    def sample_chains(self, n_sample, init_states, trace_funcs, n_process=1,
                      **kwargs):
        """Sample one or more Markov chains from given initial states.

        Performs a specified number of chain iterations (each of which may be
        composed of multiple individual Markov transitions), recording the
        outputs of functions of the sampled chain state after each iteration.
        The chains may be run in parallel across multiple independent processes
        or sequentially. In all cases all chains use independent random draws.

        Args:
            n_sample (int): Number of samples (iterations) to draw per chain.
            init_states (Iterable[ChainState] or Iterable[dict]): Initial
                chain states. Each entry can be either a `ChainState` object or
                a dictionary with entries specifying initial values for all
                state variables used by chain transition `sample` methods.
            trace_funcs (Iterable[callable]): List of functions which compute
                the variables to be recorded at each chain iteration, with each
                trace function being passed the current state and returning a
                dictionary of scalar or array values corresponding to the
                variable(s) to be stored. The keys in the returned dictionaries
                are used to index the trace arrays in the returned traces
                dictionary. If a key appears in multiple dictionaries only the
                the value corresponding to the last trace function to return
                that key will be stored.
            n_process (int or None): Number of parallel processes to run chains
                over. If set to one then chains will be run sequentially in
                otherwise a `multiprocessing.Pool` object will be used to
                dynamically assign the chains across multiple processes. If
                set to `None` then the number of processes will default to the
                output of `os.cpu_count()`.

        Kwargs:
            memmap_enabled (bool): Whether to memory-map arrays used to store
                chain data to files on disk to avoid excessive system memory
                usage for long chains and/or large chain states. The chain data
                is written to `.npy` files in the directory specified by
                `memmap_path` (or a temporary directory if not provided). These
                files persist after the termination of the function so should
                be manually deleted when no longer required. Default is to
                for memory mapping to be disabled.
            memmap_path (str): Path to directory to write memory-mapped chain
                data to. If not provided, a temporary directory will be created
                and the chain data written to files there.
            monitor_stats (Iterable[tuple(str, str)]): List of tuples of string
                key pairs, with first entry the key of a Markov transition in
                the `transitions` dict passed to the the `__init__` method and
                the second entry the key of a chain statistic that will be
                returned in the `chain_stats` dictionary. The mean over samples
                computed so far of the chain statistics associated with any
                valid key-pairs will be monitored during sampling  by printing
                as postfix to progress bar (if `tqdm` is installed).

        Returns:
            final_states (list[ChainState]): States of chains after final
                iteration. May be used to resume sampling a chain by passing as
                the initial states to a new `sample_chains` call.
            traces (dict[str, list[array]]): Dictionary of chain trace arrays.
                Values in dictionary are list of arrays of variables outputted
                by trace functions in `trace_funcs` with each array in the list
                corresponding to a single chain and the leading dimension of
                each array corresponding to the sampling (draw) index. The key
                for each value is the corresponding key in the dictionary
                returned by the trace function which computed the traced value.
            chain_stats (dict[str, dict[str, list[array]]]): Dictionary of
                chain transition statistic dictionaries. Values in outer
                dictionary are dictionaries of statistics for each chain
                transition, keyed by the string key for the transition. The
                values in each inner transition dictionary are lists of arrays
                of chain statistic values with each array in the list
                corresponding to a single chain and the leading dimension of
                each array corresponding to the sampling (draw) index. The key
                for each value is a string description of the corresponding
                integration transition statistic.
        """
        n_chain = len(init_states)
        kwargs = self.__preprocess_kwargs(kwargs)
        if RANDOMGEN_AVAILABLE:
            seed = self.rng.randint(2**64, dtype='uint64')
            rngs = [randomgen.Xorshift1024(seed).jump(i).generator
                    for i in range(n_chain)]
        else:
            seeds = (self.rng.choice(2**16, n_chain, False) * 2**16 +
                     self.rng.choice(2**16, n_chain, False))
            rngs = [np.random.RandomState(seed) for seed in seeds]
        chain_outputs = []
        shared_kwargs_list = [{
                'rng': rng,
                'n_sample': n_sample,
                'init_state': init_state,
                'trace_funcs': trace_funcs,
                'chain_index': c,
                **kwargs
            } for c, (rng, init_state) in enumerate(zip(rngs, init_states))]
        if n_process == 1:
            # Using single process therefore run chains sequentially
            for c, shared_kwargs in enumerate(shared_kwargs_list):
                final_state, traces, stats, sample_index = self._sample_chain(
                    **shared_kwargs, parallel_chains=False)
                chain_outputs.append(
                    (final_state, traces, stats, sample_index))
                if sample_index != n_sample:
                    logger.error(
                        f'Sampling manually interrupted at chain {c} iteration'
                        f' {sample_index}. Arrays containing chain traces'
                        f' and statistics computed before interruption will'
                        f' be returned.')
                    break
        else:
            # Run chains in parallel using a multiprocess(ing).Pool
            # Child processes made to ignore SIGINT signals to allow handling
            # of KeyboardInterrupts in parent process
            n_completed = 0
            pool = Pool(n_process, _ignore_sigint_initialiser)
            try:
                results = [
                    pool.apply_async(
                        self._sample_chain,
                        kwds=dict(**shared_kwargs, parallel_chains=True))
                    for shared_kwargs in shared_kwargs_list]
                for result in results:
                    chain_outputs.append(result.get())
                    n_completed += 1
            except KeyboardInterrupt:
                # Close any still running processes
                pool.terminate()
                pool.join()
                err_message = 'Sampling manually interrupted.'
                if n_completed > 0:
                    err_message += (
                        f' Data for {n_completed} completed chains will be '
                        f'returned.')
                if kwargs['memmap_enabled']:
                    err_message += (
                        f' All data recorded so far including in progress '
                        f'chains is available in directory '
                        f'{kwargs["memmap_path"]}.')
                logger.error(err_message)
        # When running parallel jobs with memory-mapping enabled, data arrays
        # returned by processes as file paths to array memory-maps therfore
        # load memory-maps objects from file before returing results
        load_memmaps = kwargs['memmap_enabled'] and n_process > 1
        return self._collate_chain_outputs(
            n_sample, chain_outputs, load_memmaps)


class HamiltonianMCMC(MarkovChainMonteCarloMethod):
    """Wrapper class for Hamiltonian Markov chain Monte Carlo (H-MCMC) methods.

    Here H-MCMC is defined as a MCMC method which augments the original target
    variable (henceforth position variable) with a momentum variable with a
    user specified conditional distribution given the position variable. In
    each chain iteration two Markov transitions leaving the resulting joint
    distribution on position and momentum variables invariant are applied -
    the momentum variables are updated in a transition which leaves their
    conditional distribution invariant (momentum transition) and then a
    trajectory in the joint space is generated by numerically integrating a
    Hamiltonian dynamic with an appropriate symplectic integrator which is
    exactly reversible, volume preserving and approximately conserves the joint
    probability density of the target-momentum state pair; one state from the
    resulting trajectory is then selected as the next joint chain state using
    an appropriate sampling scheme such that the joint distribution is left
    exactly invariant (integration transition).

    There are various options available for both the momentum transition and
    integration transition, with by default the momentum transition set to be
    independent resampling of the momentum variables from their conditional
    distribution.

    References:

      1. Duane, S., Kennedy, A.D., Pendleton, B.J. and Roweth, D., 1987.
         Hybrid Monte Carlo. Physics letters B, 195(2), pp.216-222.
      2. Neal, R.M., 2011. MCMC using Hamiltonian dynamics.
         Handbook of Markov Chain Monte Carlo, 2(11), p.2.
    """

    def __init__(self, system, rng, integration_transition,
                 momentum_transition=None):
        """
        Args:
            system: Hamiltonian system to be simulated.
            rng: Numpy RandomState random number generator instance.
            integration_transition: Markov transition operator which jointly
                updates the position and momentum components of the chain
                state by integrating the Hamiltonian dynamics of the system
                to propose new values for the state.
            momentum_transition: Markov transition operator which updates only
                the momentum component of the chain state. If set to `None` a
                transition operator which independently samples the momentum
                from its conditional distribution will be used.
        """
        self.system = system
        self.rng = rng
        if momentum_transition is None:
            momentum_transition = trans.IndependentMomentumTransition(system)
        super().__init__(rng, OrderedDict(
            momentum_transition=momentum_transition,
            integration_transition=integration_transition))

    def _preprocess_init_state(self, init_state):
        """Make sure initial state is a ChainState and has momentum."""
        if isinstance(init_state, np.ndarray):
            # If array use to set position component of new ChainState
            init_state = ChainState(pos=init_state, mom=None, dir=1)
        elif not isinstance(init_state, ChainState) or 'mom' not in init_state:
            raise TypeError(
                'init_state should be an array or `ChainState` with '
                '`mom` attribute.')
        if init_state.mom is None:
            init_state.mom = self.system.sample_momentum(init_state, self.rng)
        return init_state

    def __preprocess_kwargs(self, kwargs):
        # default to tracing only position component of state
        if 'trace_funcs' not in kwargs:
            kwargs['trace_funcs'] = [lambda state: {'pos': state.pos}]
        # if `monitor_stats` specified, expand all statistics keys to key pairs
        # with transition key set to `integration_transition`
        if 'monitor_stats' in kwargs:
            kwargs['monitor_stats'] = [
                ('integration_transition', stats_key)
                for stats_key in kwargs['monitor_stats']]
        else:
            kwargs['monitor_stats'] = [
                ('integration_transition', 'accept_prob')]
        return kwargs

    def sample_chain(self, n_sample, init_state, **kwargs):
        """Sample a Markov chain from a given initial state.

        Performs a specified number of chain iterations (each of which may be
        composed of multiple individual Markov transitions), recording the
        outputs of functions of the sampled chain state after each iteration.

        Args:
            n_sample (int): Number of samples (iterations) to draw per chain.
            init_state (ChainState or array): Initial chain state. The state
                can be either an array specifying the state position component
                or a `ChainState` instance. If an array is passed or the `mom`
                attribute of the state is not set, a momentum component will be
                independently sampled from its conditional distribution.

        Kwargs:
            trace_funcs (Iterable[callable]): List of functions which compute
                the variables to be recorded at each chain iteration, with each
                trace function being passed the current state and returning a
                dictionary of scalar or array values corresponding to the
                variable(s) to be stored. The keys in the returned dictionaries
                are used to index the trace arrays in the returned traces
                dictionary. If a key appears in multiple dictionaries only the
                the value corresponding to the last trace function to return
                that key will be stored. Default is for a single trace function
                which records the `pos` component of the state.
            memmap_enabled (bool): Whether to memory-map arrays used to store
                chain data to files on disk to avoid excessive system memory
                usage for long chains and/or large chain states. The chain data
                is written to `.npy` files in the directory specified by
                `memmap_path` (or a temporary directory if not provided). These
                files persist after the termination of the function so should
                be manually deleted when no longer required. Default is to
                for memory mapping to be disabled.
            memmap_path (str): Path to directory to write memory-mapped chain
                data to. If not provided, a temporary directory will be created
                and the chain data written to files there.
            monitor_stats (Iterable[str]): List of string keys of chain
                statistics to monitor mean of over samples computed so far
                during sampling by printing as postfix to progress bar (if
                `tqdm` is installed). Default is to print only the mean
                `accept_prob` statistic.

        Returns:
            final_state (ChainState): State of chain after final iteration. May
                be used to resume sampling a chain by passing as the initial
                state to a new `sample_chain` call.
            traces (dict[str, array]): Dictionary of chain trace arrays. Values
                in dictionary are arrays of variables outputted by trace
                functions in `trace_funcs` with leading dimension of array
                corresponding to the sampling (draw) index. The key for each
                value is the corresponding key in the dictionary returned by
                the trace function which computed the traced value.
            chain_stats (dict[str, array]): Dictionary of chain integration
                transition statistics. Values in dictionary are arrays of chain
                statistic values with the leading dimension of each array
                corresponding to the sampling (draw) index. The key for each
                value is a string description of the corresponding integration
                transition statistic.
        """
        init_state = self._preprocess_init_state(init_state)
        kwargs = self.__preprocess_kwargs(kwargs)
        final_state, traces, chain_stats = super().sample_chain(
            n_sample, init_state, **kwargs)
        chain_stats = chain_stats.get('integration_transition', {})
        return final_state, traces, chain_stats

    def sample_chains(self, n_sample, init_states, **kwargs):
        """Sample one or more Markov chains from given initial states.

        Performs a specified number of chain iterations (each of consists of a
        momentum transition and integration transition), recording the outputs
        of functions of the sampled chain state after each iteration. The
        chains may be run in parallel across multiple independent processes or
        sequentially. In all cases all chains use independent random draws.

        Args:
            n_sample (int): Number of samples (iterations) to draw per chain.
            init_states (Iterable[ChainState] or Iterable[array]): Initial
                chain states. Each state can be either an array specifying the
                state position component or a `ChainState` instance. If an
                array is passed or the `mom` attribute of the state is not set,
                a momentum component will be independently sampled from its
                conditional distribution. One chain will be run for each state
                in the iterable sequence.

        Kwargs:
            n_process (int or None): Number of parallel processes to run chains
                over. If set to one then chains will be run sequentially in
                otherwise a `multiprocessing.Pool` object will be used to
                dynamically assign the chains across multiple processes. If set
                to `None` then the number of processes will be set to the
                output of `os.cpu_count()`. Default is `n_process=1`.
            trace_funcs (Iterable[callable]): List of functions which compute
                the variables to be recorded at each chain iteration, with each
                trace function being passed the current state and returning a
                dictionary of scalar or array values corresponding to the
                variable(s) to be stored. The keys in the returned dictionaries
                are used to index the trace arrays in the returned traces
                dictionary. If a key appears in multiple dictionaries only the
                the value corresponding to the last trace function to return
                that key will be stored. Default is for a single trace function
                which records the `pos` component of the state.
            memmap_enabled (bool): Whether to memory-map arrays used to store
                chain data to files on disk to avoid excessive system memory
                usage for long chains and/or large chain states. The chain data
                is written to `.npy` files in the directory specified by
                `memmap_path` (or a temporary directory if not provided). These
                files persist after the termination of the function so should
                be manually deleted when no longer required. Default is to
                for memory mapping to be disabled.
            memmap_path (str): Path to directory to write memory-mapped chain
                data to. If not provided, a temporary directory will be created
                and the chain data written to files there.
            monitor_stats (Iterable[str]): List of string keys of chain
                statistics to monitor mean of over samples computed so far
                during sampling by printing as postfix to progress bar (if
                `tqdm` is installed). Default is to print only the mean
                `accept_prob` statistic.

        Returns:
            final_states (list[ChainState]): States of chains after final
                iteration. May be used to resume sampling a chain by passing as
                the initial states to a new `sample_chains` call.
            traces (dict[str, list[array]]): Dictionary of chain trace arrays.
                Values in dictionary are list of arrays of variables outputted
                by trace functions in `trace_funcs` with each array in the list
                corresponding to a single chain and the leading dimension of
                each array corresponding to the sampling (draw) index. The key
                for each value is the corresponding key in the dictionary
                returned by the trace function which computed the traced value.
            chain_stats (dict[str, list[array]]): Dictionary of chain
                integration transition statistics. Values in dictionary are
                lists of arrays of chain statistic values with each array in
                the list corresponding to a single chain and the leading
                dimension of each array corresponding to the sampling (draw)
                index. The key for each value is a string description of the
                corresponding integration transition statistic.
        """
        init_states = [self._preprocess_init_state(i) for i in init_states]
        kwargs = self.__preprocess_kwargs(kwargs)
        final_states, traces, chain_stats = super().sample_chains(
            n_sample, init_states, **kwargs)
        chain_stats = chain_stats.get('integration_transition', {})
        return final_states, traces, chain_stats


class StaticMetropolisHMC(HamiltonianMCMC):
    """Static integration time H-MCMC implementation with Metropolis sampling.

    In each transition a trajectory is generated by integrating the Hamiltonian
    dynamics from the current state in the current integration time direction
    for a fixed integer number of integrator steps.

    The state at the end of the trajectory with the integration direction
    negated (this ensuring the proposed move is an involution) is used as the
    proposal in a Metropolis acceptance step. The integration direction is then
    deterministically negated again irrespective of the accept decision, with
    the effect being that on acceptance the integration direction will be equal
    to its initial value and on rejection the integration direction will be
    the negation of its initial value.

    This is original proposed Hybrid Monte Carlo (often now instead termed
    Hamiltonian Monte Carlo) algorithm [1,2].

    References:

      1. Duane, S., Kennedy, A.D., Pendleton, B.J. and Roweth, D., 1987.
         Hybrid Monte Carlo. Physics letters B, 195(2), pp.216-222.
      2. Neal, R.M., 2011. MCMC using Hamiltonian dynamics.
         Handbook of Markov Chain Monte Carlo, 2(11), p.2.
    """

    def __init__(self, system, integrator, rng, n_step,
                 momentum_transition=None):
        integration_transition = trans.MetropolisStaticIntegrationTransition(
            system, integrator, n_step)
        super().__init__(system, rng, integration_transition,
                         momentum_transition)

    @property
    def n_step(self):
        return self.transitions['integration_transition'].n_step

    @n_step.setter
    def n_step(self, value):
        self.transitions['integration_transition'].n_step = value


class RandomMetropolisHMC(HamiltonianMCMC):
    """Random integration time H-MCMC with Metropolis sampling of new state.

    In each transition a trajectory is generated by integrating the Hamiltonian
    dynamics from the current state in the current integration time direction
    for a random integer number of integrator steps sampled from the uniform
    distribution on an integer interval.

    The state at the end of the trajectory with the integration direction
    negated (this ensuring the proposed move is an involution) is used as the
    proposal in a Metropolis acceptance step. The integration direction is then
    deterministically negated again irrespective of the accept decision, with
    the effect being that on acceptance the integration direction will be equal
    to its initial value and on rejection the integration direction will be
    the negation of its initial value.

    The randomisation of the number of integration steps avoids the potential
    of the chain mixing poorly due to using an integration time close to the
    period of (near) periodic systems [1,2].

    References:

      1. Neal, R.M., 2011. MCMC using Hamiltonian dynamics.
         Handbook of Markov Chain Monte Carlo, 2(11), p.2.
      2. Mackenzie, P.B., 1989. An improved hybrid Monte Carlo method.
         Physics Letters B, 226(3-4), pp.369-371.
    """

    def __init__(self, system, integrator, rng, n_step_range,
                 momentum_transition=None):
        integration_transition = trans.MetropolisRandomIntegrationTransition(
            system, integrator, n_step_range)
        super().__init__(system, rng, integration_transition,
                         momentum_transition)

    @property
    def n_step_range(self):
        return self.transitions['integration_transition'].n_step_range

    @n_step_range.setter
    def n_step_range(self, value):
        self.transitions['integration_transition'].n_step_range = value


class DynamicMultinomialHMC(HamiltonianMCMC):
    """Dynamic integration time H-MCMC with multinomial sampling of new state.

    In each transition a binary tree of states is recursively computed by
    integrating randomly forward and backward in time by a number of steps
    equal to the previous tree size [1,2] until a termination criteria on the
    tree leaves is met. The next chain state is chosen from the candidate
    states using a progressive multinomial sampling scheme [2] based on the
    relative probability densities of the different candidate states, with the
    resampling biased towards states further from the current state.

    References:

      1. Hoffman, M.D. and Gelman, A., 2014. The No-U-turn sampler:
         adaptively setting path lengths in Hamiltonian Monte Carlo.
         Journal of Machine Learning Research, 15(1), pp.1593-1623.
      2. Betancourt, M., 2017. A conceptual introduction to Hamiltonian Monte
         Carlo. arXiv preprint arXiv:1701.02434.
    """

    def __init__(self, system, integrator, rng,
                 max_tree_depth=10, max_delta_h=1000,
                 termination_criterion=trans.riemannian_no_u_turn_criterion,
                 momentum_transition=None):
        integration_transition = trans.MultinomialDynamicIntegrationTransition(
            system, integrator, max_tree_depth, max_delta_h,
            termination_criterion)
        super().__init__(system, rng, integration_transition,
                         momentum_transition)

    @property
    def max_tree_depth(self):
        return self.transitions['integration_transition'].max_tree_depth

    @max_tree_depth.setter
    def max_tree_depth(self, value):
        self.transitions['integration_transition'].max_tree_depth = value

    @property
    def max_delta_h(self):
        return self.transitions['integration_transition'].max_delta_h

    @max_delta_h.setter
    def max_delta_h(self, value):
        self.transitions['integration_transition'].max_delta_h = value



# -*- coding: utf-8 -*-
""" MCMC samplers based on simulating Hamiltonian dynamics on a manifold. """

__authors__ = 'Matt Graham'
__license__ = 'MIT'

import mici.integrators
import mici.matrices
import mici.samplers
import mici.solvers
import mici.states
import mici.systems
import mici.transitions
import mici.utils



"""Additional autograd differential operators."""

AUTOGRAD_AVAILABLE = True
try:
    from autograd import make_vjp as vjp_and_value
    from autograd.wrap_util import unary_to_nary
    from autograd.builtins import tuple as atuple
    from autograd.core import make_vjp
    from autograd.extend import vspace
    import autograd.numpy as np
except ImportError:
    AUTOGRAD_AVAILABLE = False

if AUTOGRAD_AVAILABLE:

    @unary_to_nary
    def grad_and_value(fun, x):
        """
        Returns a function that returns both gradient and value of a function.
        """
        vjp, val = make_vjp(fun, x)
        if not vspace(val).size == 1:
            raise TypeError("grad_and_value only applies to real scalar-output"
                            " functions.")
        return vjp(vspace(val).ones()), val

    @unary_to_nary
    def jacobian_and_value(fun, x):
        """
        Returns a function that returns both the Jacobian and value of a
        function.

        Assumes that the function `fun` broadcasts along the first dimension of
        the input being differentiated with respect to such that a batch of
        outputs can be computed concurrently for a batch of inputs.
        """
        val = fun(x)
        v_vspace = vspace(val)
        x_vspace = vspace(x)
        x_rep = np.tile(x, (v_vspace.size,) + (1,) * x_vspace.ndim)
        vjp_rep, _ = make_vjp(fun, x_rep)
        jacobian_shape = v_vspace.shape + x_vspace.shape
        basis_vectors = np.array([b for b in v_vspace.standard_basis()])
        jacobian = vjp_rep(basis_vectors)
        return np.reshape(jacobian, jacobian_shape), val

    @unary_to_nary
    def mhp_jacobian_and_value(fun, x):
        """
        Returns a function that returns MHP, Jacobian and value of a function.

        For a vector-valued function `fun` the matrix-Hessian-product (MHP) is
        here defined as a function of a matrix `m` corresponding to

            mhp(m) = sum(m[:, :, None] * h[:, :, :], axis=(0, 1))

        where `h` is the vector-Hessian of `f = fun(x)` wrt `x` i.e. the rank-3
        tensor of second-order partial derivatives of the vector-valued
        function, such that

            h[i, j, k] = (d**2 f[i]) / (dx[j] * dx[k])

        Assumes that the function `fun` broadcasts along the first dimension of
        the input being differentiated with respect to such that a batch of
        outputs can be computed concurrently for a batch of inputs.
        """
        mhp, (jacob, val) = make_vjp(
            lambda x: atuple(jacobian_and_value(fun)(x)), x)
        return lambda m: mhp((m, vspace(val).zeros())), jacob, val

    @unary_to_nary
    def hessian_grad_and_value(fun, x):
        """
        Returns a function that returns the Hessian, gradient and value of a
        function.

        Assumes that the function `fun` broadcasts along the first dimension of
        the input being differentiated with respect to such that a batch of
        outputs can be computed concurrently for a batch of inputs.
        """
        def grad_fun(x):
            vjp, val = make_vjp(fun, x)
            return vjp(vspace(val).ones()), val
        x_vspace = vspace(x)
        x_rep = np.tile(x, (x_vspace.size,) + (1,) * x_vspace.ndim)
        vjp_grad, (grad, val) = make_vjp(lambda x: atuple(grad_fun(x)), x_rep)
        hessian_shape = x_vspace.shape + x_vspace.shape
        basis_vectors = np.array([b for b in x_vspace.standard_basis()])
        hessian = vjp_grad((basis_vectors, vspace(val).zeros()))
        return np.reshape(hessian, hessian_shape), grad[0], val[0]

    @unary_to_nary
    def mtp_hessian_grad_and_value(fun, x):
        """
        Returns a function that returns MTP, Jacobian and value of a function.

        For a scalar-valued function `fun` the matrix-Tressian-product (MTP) is
        here defined as a function of a matrix `m` corresponding to

            mtp(m) = sum(m[:, :] * t[:, :, :], axis=(-1, -2))

        where `t` is the 'Tressian' of `f = fun(x)` wrt `x` i.e. the rank-3
        tensor of third-order partial derivatives of the scalar-valued function
        such that

            t[i, j, k] = d**3 f / (dx[i] * dx[j] * dx[k])

        Assumes that the function `fun` broadcasts along the first dimension of
        the input being differentiated with respect to such that a batch of
        outputs can be computed concurrently for a batch of inputs.
        """
        mtp, (hessian, grad, val) = make_vjp(
            lambda x: atuple(hessian_grad_and_value(fun)(x)), x)
        return (
            lambda m: mtp((m, vspace(grad).zeros(), vspace(val).zeros())),
            hessian, grad, val)



"""Solvers for non-linear systems of equations for implicit integrators."""

from mici.errors import ConvergenceError
import numpy as np


def euclidean_norm(vct):
    """Calculate the Euclidean (L-2) norm of a vector."""
    return np.sum(vct**2)**0.5


def maximum_norm(vct):
    """Calculate the maximum (L-infinity) norm of a vector."""
    return np.max(abs(vct))


def solve_fixed_point_direct(func, x0, tol=1e-9, max_iters=100,
                             norm=maximum_norm):
    """Solve fixed point equation `func(x) = x` using direct iteration.

    Args:
        func: Single argument function to find fixed point of.
        x0: Initial state (function argument).
        tol: Convergence tolerance - terminates when `norm(func(x) - x) < tol`.
        max_iters: Maximum number of iterations before raising exception.
        norm: Vector norm to use to assess convergence.

    Returns:
        Solution to fixed point equation with `norm(func(x) - x) < tol`.

    Raises:
        ConvergenceError if not converged within `max_iters` iterations.
    """
    for i in range(max_iters):
        x = func(x0)
        if any(np.isnan(x)) or any(np.isinf(x)):
            raise ConvergenceError(
                f'Fixed point iteration diverged on iteration {i}.')
        error = norm(x - x0)
        if error < tol:
            return x
        x0 = x
    raise ConvergenceError(
        f'Fixed point iteration did not converge. Last error {error:.1e}.')


def solve_fixed_point_steffensen(func, x0, tol=1e-9, max_iters=100,
                                 norm=maximum_norm):
    """Solve fixed point equation `func(x) = x` using Steffensen's method.

    Steffennsen's method [1] achieves quadratic convergence but at the cost of
    two function evaluations per iteration so for functions where convergence
    is achieved in a small number of iterations, direct iteration may be
    cheaper.

    [1] : https://en.wikipedia.org/wiki/Steffensen%27s_method

    Args:
        func: Single argument function to find fixed point of.
        x0: Initial state (function argument).
        tol: Convergence tolerance - terminates when `norm(func(x) - x) < tol`.
        max_iters: Maximum number of iterations before raising exception.
        norm: Vector norm to use to assess convergence.

    Returns:
        Solution to fixed point equation with `norm(func(x) - x) < tol`.

    Raises:
        ConvergenceError if not converged within `max_iters` iterations.
    """
    for i in range(max_iters):
        x1 = func(x0)
        x2 = func(x1)
        x = x0 - (x1 - x0)**2 / (x2 - 2 * x1 + x0)
        if any(np.isnan(x)) or any(np.isinf(x)):
            raise ConvergenceError(
                f'Fixed point iteration diverged on iteration {i}.')
        error = norm(x - x0)
        if error < tol:
            return x
        x0 = x
    raise ConvergenceError(
        f'Fixed point iteration did not converge. Last error {error:.1e}.')


def retract_onto_manifold_quasi_newton(
        state, state_prev, dt, system, convergence_tol=1e-9,
        divergence_tol=1e10, max_iters=50, norm=maximum_norm):
    mu = np.zeros_like(state.pos)
    jacob_constr_prev = system.jacob_constr(state_prev)
    dh2_flow_pos_dmom, dh2_flow_mom_dmom = system.dh2_flow_dmom(dt)
    inv_jacob_constr_inner_product = system.jacob_constr_inner_product(
        jacob_constr_prev, dh2_flow_pos_dmom).inv
    for i in range(max_iters):
        try:
            constr = system.constr(state)
            error = norm(constr)
            if error > divergence_tol or np.isnan(error):
                raise ConvergenceError(
                    'Quasi-Newton iteration diverged. Last error {err:.1e}.')
            elif error < convergence_tol:
                state.mom -= dh2_flow_mom_dmom @ mu
                return state
            delta_mu = jacob_constr_prev.T @ (
                inv_jacob_constr_inner_product @ constr)
            mu += delta_mu
            state.pos -= dh2_flow_pos_dmom @ delta_mu
        except ValueError as e:
            # Make robust to inf/nan values in intermediate linear algebra ops
            raise ConvergenceError(
                f'ValueError during Quasi-Newton iteration ({e}).')
    raise ConvergenceError(
        f'Quasi-Newton iteration did not converge. Last error {error:.1e}.')


def retract_onto_manifold_newton(
        state, state_prev, dt, system, convergence_tol=1e-9,
        divergence_tol=1e10, max_iters=50, norm=maximum_norm):
    mu = np.zeros_like(state.pos)
    jacob_constr_prev = system.jacob_constr(state_prev)
    dh2_flow_pos_dmom, dh2_flow_mom_dmom = system.dh2_flow_dmom(dt)
    for i in range(max_iters):
        try:
            jacob_constr = system.jacob_constr(state)
            constr = system.constr(state)
            error = norm(constr)
            if error > divergence_tol or np.isnan(error):
                raise ConvergenceError(
                    'Newton iteration diverged. Last error {err:.1e}.')
            if error < convergence_tol:
                state.mom -= dh2_flow_mom_dmom @ mu
                return state
            delta_mu = jacob_constr_prev.T @ (
                system.jacob_constr_inner_product(
                    jacob_constr, dh2_flow_pos_dmom, jacob_constr_prev).inv @
                constr)
            mu += delta_mu
            state.pos -= dh2_flow_pos_dmom @ delta_mu
        except ValueError as e:
            # Make robust to inf/nan values in intermediate linear algebra ops
            raise ConvergenceError(
                f'ValueError during Newton iteration ({e}).')
    raise ConvergenceError(
        f'Newton iteration did not converge. Last error {error:.1e}.')



"""Utility functions and classes."""

import numpy as np
from math import log, exp, log1p, expm1, inf, nan
import types
import sys
import inspect
import mici


try:

    import tqdm
    import logging

    class TqdmHandler(logging.StreamHandler):
        """Simple log handler which uses tqdm write method."""

        def __init__(self):
            super().__init__()

        def emit(self, record):
            msg = self.format(record)
            tqdm.tqdm.write(msg)

    def setup_tqdm_logger():
        """Returns a logger which redirects log output to tqdm.write."""
        logger = logging.getLogger()
        handler = TqdmHandler()
        logger.addHandler(handler)
        return logger

except ImportError:
    pass


try:

    import arviz

    def convert_to_arviz_inference_data(
            traces, chain_stats, sample_stats_key=None):
        """Wrap chain outputs in an `arviz.InferenceData` container object.

        The `traces` and `chain_stats` arguments should correspond to a
        multiple-chain sampler output i.e. the returned values from a
        `sample_chains` call.

        Args:
            traces (dict[str, list[array]]): Trace arrays, with one entry per
                function in `trace_funcs` passed to sampler method. Each entry
                consists of a list of arrays, one per chain, with the first
                axes of the arrays corresponding to the sampling (draw) index.
            chain_stats (dict[str, list[array]]): Chain integration transition
                statistics as a dictionary with string keys describing the
                statistics recorded and values corresponding to a list of
                arrays with one array per chain and the first axis of the
                arrays corresponding to the sampling index.
            sample_stats_key (str): Optional. Key of transition in
                `chain_stats` to use the recorded statistics of to populate the
                `sampling_stats` group in the returned `InferenceData` object.

        Returns:
            arvix.InferenceData:
                An arviz data container with groups `posterior` and
                'sample_stats', both of instances of `xarray.Dataset`. The
                `posterior` group corresponds to the chain variable traces
                provides in the `traces` argument and the `sample_stats`
                group corresponds to the chain transitions statistics passed
                in the `chain_stats` argument (if multiple transition
                statistics dictionaries are present the `sample_stats_key`
                argument should be specified to indicate which to use).
        """
        if (sample_stats_key is not None and
                sample_stats_key not in chain_stats):
            raise ValueError(
                f'Specified `sample_stats_key` ({sample_stats_key}) does '
                f'not match any transition in `chain_stats`.')
        if sample_stats_key is not None:
            return arviz.InferenceData(
                posterior=arviz.dict_to_dataset(traces, library=mici),
                sample_stats=arviz.dict_to_dataset(
                    chain_stats[sample_stats_key], library=mici))
        elif not isinstance(next(iter(chain_stats.values())), dict):
            # chain_stats dictionary value not another dictionary therefore
            # assume corresponds to statistics for a single transition
            return arviz.InferenceData(
                posterior=arviz.dict_to_dataset(traces, library=mici),
                sample_stats=arviz.dict_to_dataset(chain_stats, library=mici))
        elif len(chain_stats) == 1:
            # single transtition statistics dictionary in chain_stats therefore
            # unambiguous to set sample_stats
            return arviz.InferenceData(
                posterior=arviz.dict_to_dataset(traces, library=mici),
                sample_stats=arviz.dict_to_dataset(
                    chain_stats.popitem()[1], library=mici))
        else:
            raise ValueError(
                '`sample_stats_key` must be specified as `chain_stats` '
                'contains multiple transtitiion statistics dictionaries.')

except ImportError:
    pass


def get_valid_filename(string):
    """Generate a valid filename from a string.

    Strips all characters which are not alphanumeric or a period (.), dash (-)
    or underscore (_).

    Based on https://stackoverflow.com/a/295146/4798943
    """
    return ''.join(c for c in string if (c.isalnum() or c in '._- '))


def get_size(obj, seen=None):
    """Recursively finds size of objects in bytes.

    Original source: https://github.com/bosswissam/pysize

    MIT License

    Copyright (c) [2018] [Wissam Jarjoui]

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if hasattr(obj, '__dict__'):
        for cls in obj.__class__.__mro__:
            if '__dict__' in cls.__dict__:
                d = cls.__dict__['__dict__']
                if (inspect.isgetsetdescriptor(d) or
                        inspect.ismemberdescriptor(d)):
                    size += get_size(obj.__dict__, seen)
                break
    if isinstance(obj, dict):
        size += sum((get_size(v, seen) for v in obj.values()))
        size += sum((get_size(k, seen) for k in obj.keys()))
    elif (hasattr(obj, '__iter__') and not
            isinstance(obj, (str, bytes, bytearray))):
        size += sum((get_size(i, seen) for i in obj))

    if hasattr(obj, '__slots__'):  # can have __slots__ with __dict__
        size += sum(get_size(getattr(obj, s), seen)
                    for s in obj.__slots__ if hasattr(obj, s))

    return size


def inherit_docstrings(cls):
    """Class decorator to inherit parent class docstrings if not specified.

    Taken from https://stackoverflow.com/a/38601305/4798943.
    """
    for name, func in vars(cls).items():
        if isinstance(func, types.FunctionType) and not func.__doc__:
            for parent in cls.__bases__:
                parfunc = getattr(parent, name, None)
                if parfunc and getattr(parfunc, '__doc__', None):
                    func.__doc__ = parfunc.__doc__
                    break
        elif isinstance(func, property) and not func.fget.__doc__:
            for parent in cls.__bases__:
                parprop = getattr(parent, name, None)
                if parprop and getattr(parprop.fget, '__doc__', None):
                    newprop = property(fget=func.fget,
                                       fset=func.fset,
                                       fdel=func.fdel,
                                       doc=parprop.fget.__doc__)
                    setattr(cls, name, newprop)
                    break

    return cls


LOG_2 = log(2.)


def log1p_exp(val):
    if val > 0.:
        return val + log1p(exp(-val))
    else:
        return log1p(exp(val))


def log1m_exp(val):
    if val >= 0.:
        return nan
    elif val > LOG_2:
        return log(-expm1(val))
    else:
        return log1p(-exp(a))


def log_sum_exp(val1, val2):
    if val1 > val2:
        return val1 + log1p_exp(val2 - val1)
    else:
        return val2 + log1p_exp(val1 - val2)


def log_diff_exp(val1, val2):
    if val1 < val2:
        return nan
    elif val1 == val2:
        return -inf
    else:
        return val1 + log1m_exp(val2 - val1)


class LogRepFloat(object):
    """Numerically stable logarithmic representation of positive float values.

    Stores logarithm of value and overloads arithmetic operators to use more
    numerically stable implementations where possible.
    """

    def __init__(self, val=None, log_val=None):
        if log_val is None:
            if val is None:
                raise ValueError('One of val or log_val must be specified.')
            elif val > 0:
                self.log_val = log(val)
            elif val == 0.:
                self.log_val = -inf
            else:
                raise ValueError('val must be non-negative.')
        else:
            if val is not None:
                raise ValueError('Specify only one of val and log_val.')
            else:
                self.log_val = log_val

    @property
    def val(self):
        try:
            return exp(self.log_val)
        except OverflowError:
            return inf

    def __add__(self, other):
        if isinstance(other, LogRepFloat):
            return LogRepFloat(
                log_val=log_sum_exp(self.log_val, other.log_val))
        else:
            return self.val + other

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        if isinstance(other, LogRepFloat):
            self.log_val = log_sum_exp(self.log_val, other.log_val)
        else:
            self.log_val = log_sum_exp(self.log_val, log(other))
        return self

    def __sub__(self, other):
        if isinstance(other, LogRepFloat):
            if self.log_val >= other.log_val:
                return LogRepFloat(
                    log_val=log_diff_exp(self.log_val, other.log_val))
            else:
                return self.val - other.val
        else:
            return self.val - other

    def __mul__(self, other):
        if isinstance(other, LogRepFloat):
            return LogRepFloat(log_val=self.log_val + other.log_val)
        else:
            return self.val * other

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, LogRepFloat):
            return LogRepFloat(log_val=self.log_val - other.log_val)
        else:
            return self.val / other

    def __neg__(self):
        return -self.val

    def __eq__(self, other):
        if isinstance(other, LogRepFloat):
            return self.log_val == other.log_val
        else:
            return self.val == other

    def __ne__(self, other):
        if isinstance(other, LogRepFloat):
            return self.log_val != other.log_val
        else:
            return self.val != other

    def __lt__(self, other):
        if isinstance(other, LogRepFloat):
            return self.log_val < other.log_val
        else:
            return self.val < other

    def __gt__(self, other):
        if isinstance(other, LogRepFloat):
            return self.log_val > other.log_val
        else:
            return self.val > other

    def __le__(self, other):
        if isinstance(other, LogRepFloat):
            return self.log_val <= other.log_val
        else:
            return self.val <= other

    def __ge__(self, other):
        if isinstance(other, LogRepFloat):
            return self.log_val >= other.log_val
        else:
            return self.val >= other

    def __str__(self):
        return str(self.val)

    def __repr__(self):
        return 'LogRepFloat(val={0})'.format(self.val)



import setuptools

setuptools.setup(
    name='mici',
    version='0.1.0-r1',
    author='Matt Graham',
    description=(
        'MCMC samplers based on simulating Hamiltonian dynamics on a manifold'
    ),
    long_description=(
        'Mici is a Python package providing implementations of Markov chain '
        'Monte Carlo (MCMC) methods for approximate inference in probabilistic'
        ' models, with a particular focus on MCMC methods based on simulating '
        'Hamiltonian dynamics on a manifold.'
    ),
    url='https://github.com/matt-graham/mici.git',
    project_urls={
        'Documentation': 'https://matt-graham.github.io/mici/docs'
    },
    packages=['mici'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers'
    ],
    keywords='inference sampling MCMC HMC',
    license='MIT',
    license_file='LICENSE',
    install_requires=['numpy>=1.15', 'scipy>=1.1'],
    python_requires='>=3.6'
)



"""Symplectic integrators for simulation of Hamiltonian dynamics."""

from mici.errors import NonReversibleStepError
from mici.solvers import (maximum_norm, solve_fixed_point_direct,
                          retract_onto_manifold_quasi_newton)


class ExplicitLeapfrogIntegrator(object):
    """
    Leapfrog integrator for Hamiltonian systems with tractable component flows.

    The Hamiltonian function is assumed to be expressible as the sum of two
    analytically tractable components for which the corresponding Hamiltonian
    flows can be exactly simulated. Specifically it is assumed that the
    Hamiltonian function `h` takes the form

        h(pos, mom) = h1(pos) + h2(pos, mom)

    where `pos` and `mom` are the position and momentum variables respectively,
    and `h1(pos)` and `h2(pos, mom)` Hamiltonian components for which the exact
    flows can be computed.
    """

    def __init__(self, system, step_size):
        if not hasattr(system, 'h1_flow') or not hasattr(system, 'h2_flow'):
            raise ValueError(
                'Explicit leapfrog integrator can only be used for systems '
                'with explicit `h1_flow` and `h2_flow` Hamiltonian component '
                'flow maps. For systems in which only `h1_flow` is available '
                'the `ImplicitLeapfrogIntegrator` class may be used instead.')
        self.system = system
        self.step_size = step_size

    def step(self, state):
        dt = state.dir * self.step_size
        state = state.copy()
        self.system.h1_flow(state, 0.5 * dt)
        self.system.h2_flow(state, dt)
        self.system.h1_flow(state, 0.5 * dt)
        return state


LeapfrogIntegrator = ExplicitLeapfrogIntegrator


class ImplicitLeapfrogIntegrator(object):
    """
    Implicit leapfrog integrator for Hamiltonian with non-separable component.

    The Hamiltonian function is assumed to take the form

        h(pos, mom) = h1(pos) + h2(pos, mom)

    where `pos` and `mom` are the position and momentum variables respectively,
    and `h2` is a non-separable function of the position and momentum variables
    and for which exact simulation of the correspond Hamiltonian flow is not
    possible. A pair of implicit component updates are used to approximate the
    flow due to the `h2` Hamiltonian component, with a fixed-point iteration
    used to solve the non-linear system of equations.
    """

    def __init__(self, system, step_size, reverse_check_tol=1e-8,
                 reverse_check_norm=maximum_norm,
                 fixed_point_solver=solve_fixed_point_direct,
                 **fixed_point_solver_kwargs):
        self.system = system
        self.step_size = step_size
        self.reverse_check_tol = reverse_check_tol
        self.reverse_check_norm = maximum_norm
        self.fixed_point_solver = fixed_point_solver
        self.fixed_point_solver_kwargs = fixed_point_solver_kwargs

    def solve_fixed_point(self, fixed_point_func, x_init):
        return self.fixed_point_solver(
            fixed_point_func, x_init, **self.fixed_point_solver_kwargs)

    def step_a(self, state, dt):
        self.system.h1_flow(state, dt)

    def step_b_fwd(self, state, dt):
        def fixed_point_func(mom):
            state.mom = mom
            return mom_init - dt * self.system.dh2_dpos(state)
        mom_init = state.mom
        state.mom = self.solve_fixed_point(fixed_point_func, mom_init)

    def step_b_adj(self, state, dt):
        mom_init = state.mom.copy()
        state.mom -= dt * self.system.dh2_dpos(state)
        state_back = state.copy()
        self.step_b_fwd(state_back, -dt)
        rev_diff = self.reverse_check_norm(state_back.mom - mom_init)
        if rev_diff > self.reverse_check_tol:
            raise NonReversibleStepError(
                f'Non-reversible step. Distance between initial and '
                f'forward-backward integrated momentums = {rev_diff:.1e}.')

    def step_c_fwd(self, state, dt):
        pos_init = state.pos.copy()
        state.pos += dt * self.system.dh2_dmom(state)
        state_back = state.copy()
        self.step_c_adj(state_back, -dt)
        rev_diff = self.reverse_check_norm(state_back.pos - pos_init)
        if rev_diff > self.reverse_check_tol:
            raise NonReversibleStepError(
                f'Non-reversible step. Distance between initial and '
                f'forward-backward integrated positions = {rev_diff:.1e}.')

    def step_c_adj(self, state, dt):
        def fixed_point_func(pos):
            state.pos = pos
            return pos_init + dt * self.system.dh2_dmom(state)
        pos_init = state.pos
        state.pos = self.solve_fixed_point(fixed_point_func, pos_init)

    def step(self, state):
        dt = 0.5 * state.dir * self.step_size
        state = state.copy()
        self.step_a(state, dt)
        self.step_b_fwd(state, dt)
        self.step_c_fwd(state, dt)
        self.step_c_adj(state, dt)
        self.step_b_adj(state, dt)
        self.step_a(state, dt)
        return state


class ConstrainedLeapfrogIntegrator(object):
    """
    Leapfrog integrator for constrained Hamiltonian systems.

    The Hamiltonian function is assumed to be expressible as the sum of two
    components for which the corresponding (unconstrained) Hamiltonian flows
    can be exactly simulated. Specifically it is assumed that the Hamiltonian
    function `h` takes the form

        h(pos, mom) = h1(pos) + h2(pos, mom)

    where `pos` and `mom` are the position and momentum variables respectively,
    and `h1(pos)` and `h2(pos, mom)` Hamiltonian components for which the exact
    flows can be computed.

    The system is assumed to be additionally subject to a set of holonomic
    constraints on the position component of the state i.e. that all valid
    states must satisfy

        all(constr(pos) == 0)

    for some vector constraint function `constr`, with the set of positions
    satisfying the constraints implicitly defining a manifold. There is also
    a corresponding constraint implied on the momentum variables which can
    be derived by differentiating the above with respect to time and using
    that under the Hamiltonian dynamics the time derivative of the position
    is equal to the negative derivative of the Hamiltonian function with
    respect to the momentum

        all(jacob_constr(pos) @ dh2_dmom(mom) == 0)

    The set of momentum variables satisfying the above for given position
    variables is termed the cotangent space of the manifold (at a position),
    and the set of (position, momentum) pairs for which the position is on the
    constraint manifold and the momentum in the corresponding cotangent space
    is termed the cotangent bundle.

    The integrator exactly preserves these constraints at all steps, such that
    if an initial position momentum pair `(pos, mom)` are in the cotangent
    bundle, the corresponding pair after calling the `step` method of the
    integrator will also be in the cotangent bundle.
    """

    def __init__(self, system, step_size, n_inner_step=1,
                 reverse_check_tol=1e-8, reverse_check_norm=maximum_norm,
                 retraction_solver=retract_onto_manifold_quasi_newton,
                 retraction_solver_kwargs=None):
        self.system = system
        self.step_size = step_size
        self.n_inner_step = n_inner_step
        self.reverse_check_tol = reverse_check_tol
        self.reverse_check_norm = reverse_check_norm
        self.retraction_solver = retraction_solver
        if retraction_solver_kwargs is None:
            retraction_solver_kwargs = {}
        self.retraction_solver_kwargs = retraction_solver_kwargs

    def retract_onto_manifold(self, state, state_prev, dt):
        self.retraction_solver(state, state_prev, dt, self.system,
                               **self.retraction_solver_kwargs)

    def project_onto_cotangent_space(self, state):
        self.system.project_onto_cotangent_space(state.mom, state)

    def step_a(self, state, dt):
        self.system.h1_flow(state, dt)
        self.project_onto_cotangent_space(state)

    def step_b(self, state, dt):
        dt_i = dt / self.n_inner_step
        for i in range(self.n_inner_step):
            state_prev = state.copy()
            self.system.h2_flow(state, dt_i)
            self.retract_onto_manifold(state, state_prev, dt_i)
            self.project_onto_cotangent_space(state)
            state_back = state.copy()
            self.system.h2_flow(state_back, -dt_i)
            self.retract_onto_manifold(state_back, state, -dt_i)
            rev_diff = self.reverse_check_norm(state_back.pos - state_prev.pos)
            if rev_diff > self.reverse_check_tol:
                raise NonReversibleStepError(
                    f'Non-reversible step. Distance between initial and '
                    f'forward-backward integrated positions = {rev_diff:.1e}.')

    def step(self, state):
        dt = state.dir * self.step_size
        state = state.copy()
        self.step_a(state, 0.5 * dt)
        self.step_b(state, dt)
        self.step_a(state, 0.5 * dt)
        return state
